"""
Module to helper with the parsing of link reference definitions.
"""

import logging
from typing import Optional, Tuple

from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.general.parser_state import ParserState
from pymarkdown.general.tab_helper import TabHelper
from pymarkdown.inline.inline_backslash_helper import InlineBackslashHelper
from pymarkdown.links.link_parse_helper import LinkParseHelper
from pymarkdown.links.link_reference_info import LinkReferenceInfo
from pymarkdown.links.link_reference_titles import LinkReferenceTitles
from pymarkdown.links.link_reference_tuple import LinkReferenceDefinitionTuple

POGGER = ParserLogger(logging.getLogger(__name__))


# pylint: disable=too-few-public-methods
class LinkReferenceDefinitionParseHelper:
    """
    Class to helper with the parsing of link reference definitions.
    """

    __lrd_start_character = "["

    # pylint: disable=too-many-locals
    @staticmethod
    def parse_link_reference_definition(
        parser_state: ParserState,
        line_to_parse: str,
        start_index: int,
        extracted_whitespace: str,
        is_blank_line: bool,
        remaining_line_to_parse: str,
        was_started: bool,
    ) -> Tuple[bool, int, Optional[LinkReferenceDefinitionTuple]]:
        """
        Handle the parsing of what appears to be a link reference definition.
        """

        did_start = LinkReferenceDefinitionParseHelper.__is_link_reference_definition(
            parser_state, line_to_parse, start_index, extracted_whitespace
        )
        if not did_start:
            POGGER.debug("BAIL")
            return False, -1, None

        new_index = -1
        (
            keep_going,
            new_index,
            collected_destination,
        ) = LinkParseHelper.extract_link_label(
            parser_state.parse_properties, line_to_parse, start_index + 1
        )
        if keep_going:
            (
                keep_going,
                new_index,
                inline_link,
                _,
                line_destination_whitespace,
                inline_raw_link,
            ) = LinkParseHelper.extract_link_destination(
                parser_state.parse_properties, line_to_parse, new_index, is_blank_line
            )
        if keep_going:
            (
                keep_going,
                new_index,
                inline_title,
                _,
                line_title_whitespace,
                inline_raw_title,
            ) = LinkParseHelper.extract_link_title(
                parser_state.parse_properties, line_to_parse, new_index, is_blank_line
            )
        else:
            inline_title = ""
        if keep_going:
            (
                keep_going,
                new_index,
                end_whitespace,
            ) = LinkReferenceDefinitionParseHelper.__verify_link_definition_end(
                line_to_parse, new_index
            )
        if keep_going:
            assert (
                collected_destination is not None
            ), "if extract_link_label returned true, this must be defined."
            normalized_destination = LinkParseHelper.normalize_link_label(
                collected_destination
            )
            if not normalized_destination:
                keep_going, new_index = False, -1
        else:
            normalized_destination = None
        return (
            LinkReferenceDefinitionParseHelper.__create_lrd_token(
                new_index,
                collected_destination,
                normalized_destination,
                line_destination_whitespace,
                inline_link,
                inline_raw_link,
                inline_title,
                inline_raw_title,
                line_title_whitespace,
                end_whitespace,
            )
            if keep_going
            else (keep_going, new_index, None)
        )

    # pylint: enable=too-many-locals

    @staticmethod
    def is_link_reference_definition(
        parser_state: ParserState,
        line_to_parse: str,
        start_index: int,
        extracted_whitespace: str,
    ) -> bool:
        return LinkReferenceDefinitionParseHelper.__is_link_reference_definition(
            parser_state, line_to_parse, start_index, extracted_whitespace
        )

    @staticmethod
    def __is_link_reference_definition(
        parser_state: ParserState,
        line_to_parse: str,
        start_index: int,
        extracted_whitespace: str,
    ) -> bool:
        """
        Determine whether or not we have the start of a link reference definition.
        """

        if parser_state.token_stack[-1].is_paragraph:
            return False

        POGGER.debug(
            "__is_link_reference_definition - extracted_whitespace:>:$:<",
            extracted_whitespace,
        )
        POGGER.debug(
            "__is_link_reference_definition - line_to_parse:>:$:<", line_to_parse
        )
        if (
            TabHelper.is_length_less_than_or_equal_to(extracted_whitespace, 3)
        ) and ParserHelper.is_character_at_index_one_of(
            line_to_parse,
            start_index,
            LinkReferenceDefinitionParseHelper.__lrd_start_character,
        ):
            POGGER.debug("__is_link_reference_definition - potential")
            remaining_line, continue_with_lrd = line_to_parse[start_index + 1 :], True
            if (
                remaining_line
                and remaining_line[-1] == InlineBackslashHelper.backslash_character
            ):
                remaining_line_size, start_index, found_index = (
                    len(remaining_line),
                    0,
                    remaining_line.find(
                        InlineBackslashHelper.backslash_character, start_index
                    ),
                )
                POGGER.debug(">>$<<$", remaining_line, remaining_line_size)
                POGGER.debug(">>$<<$", remaining_line, start_index)
                POGGER.debug(">>$<<", found_index)
                while found_index != -1 and found_index < (remaining_line_size - 1):
                    start_index = found_index + 2
                    POGGER.debug(">>$<<$", remaining_line, start_index)
                    found_index = remaining_line.find(
                        InlineBackslashHelper.backslash_character, start_index
                    )
                    POGGER.debug(">>$<<", found_index)
                POGGER.debug(">>>>>>>$<<", found_index)
                continue_with_lrd = found_index != remaining_line_size - 1
            return continue_with_lrd
        return False

    # pylint: disable=too-many-arguments
    @staticmethod
    def __create_lrd_token(
        new_index: int,
        collected_destination: Optional[str],
        normalized_destination: Optional[str],
        line_destination_whitespace: Optional[str],
        inline_link: Optional[str],
        inline_raw_link: Optional[str],
        inline_title: Optional[str],
        inline_raw_title: Optional[str],
        line_title_whitespace: Optional[str],
        end_whitespace: Optional[str],
    ) -> Tuple[bool, int, Optional[LinkReferenceDefinitionTuple]]:
        POGGER.debug(
            ">>collected_destination(normalized)>>$",
            normalized_destination,
        )

        if (
            not inline_title
            and line_title_whitespace
            and line_title_whitespace[-1] == ParserHelper.newline_character
        ):
            line_title_whitespace = line_title_whitespace[:-1]

        POGGER.debug(">>inline_link>>$<<", inline_link)
        POGGER.debug(">>inline_title>>$<<", inline_title)
        parsed_lrd_tuple = LinkReferenceDefinitionTuple(
            normalized_destination,
            LinkReferenceTitles(inline_link, inline_title),
            LinkReferenceInfo(
                collected_destination,
                line_destination_whitespace,
                inline_raw_link,
                line_title_whitespace,
                inline_raw_title,
                end_whitespace,
            ),
        )
        return True, new_index, parsed_lrd_tuple

    # pylint: enable=too-many-arguments

    @staticmethod
    def __verify_link_definition_end(
        line_to_parse: str, new_index: int
    ) -> Tuple[bool, int, Optional[str]]:
        """
        Verify that the link reference definition's ends properly.
        """

        POGGER.debug("look for EOL-ws>>$<<", line_to_parse[new_index:])
        new_index, ex_ws = ParserHelper.extract_ascii_whitespace_verified(
            line_to_parse, new_index
        )
        POGGER.debug("look for EOL>>$<<", line_to_parse[new_index:])
        if new_index < len(line_to_parse):
            POGGER.debug(">> characters left at EOL, bailing")
            return False, -1, None
        return True, new_index, ex_ws


# pylint: enable=too-few-public-methods



"""
Module to implement a plugin that ensures consistent spacing after the list markers.
"""

from typing import Dict, List, Optional, cast

from pymarkdown.plugin_manager.plugin_details import (
    PluginDetailsV2,
    PluginDetailsV3,
    QueryConfigItem,
)
from pymarkdown.plugin_manager.plugin_scan_context import PluginScanContext
from pymarkdown.plugin_manager.rule_plugin import RulePlugin
from pymarkdown.plugins.utils.list_tracker import ListTracker
from pymarkdown.tokens.list_start_markdown_token import ListStartMarkdownToken
from pymarkdown.tokens.markdown_token import EndMarkdownToken, MarkdownToken


# pylint: disable=too-many-instance-attributes
class RuleMd030(RulePlugin):
    """
    Class to implement a plugin that ensures consistent spacing after the list markers.
    """

    def __init__(self) -> None:
        super().__init__()
        self.__ul_single: int = -1
        self.__ul_multi: int = -1
        self.__ol_single: int = -1
        self.__ol_multi: int = -1
        self.__list_stack: List[MarkdownToken] = []
        self.__list_tokens: List[List[MarkdownToken]] = []
        self.__current_list_parent: Optional[MarkdownToken] = None
        self.__paragraph_count_map: Dict[str, int] = {}
        # self.__debug = False
        self.__frank = ListTracker()

    def get_details(self) -> PluginDetailsV2:
        """
        Get the details for the plugin.
        """
        return PluginDetailsV3(
            plugin_name="list-marker-space",
            plugin_id="MD030",
            plugin_enabled_by_default=True,
            plugin_description="Spaces after list markers",
            plugin_version="0.6.0",
            plugin_url="https://pymarkdown.readthedocs.io/en/latest/plugins/rule_md030.md",
            plugin_configuration="ul_single,ol_single,ul_multi,ol_multi",
            plugin_supports_fix=True,
        )

    @classmethod
    def __validate_minimum(cls, found_value: int) -> None:
        if found_value < 1:
            raise ValueError("Allowable values are any integer greater than 0.")

    def initialize_from_config(self) -> None:
        """
        Event to allow the plugin to load configuration information.
        """
        self.__ul_single = self.plugin_configuration.get_integer_property_with_default(
            "ul_single",
            1,
            valid_value_fn=self.__validate_minimum,
        )
        self.__ul_multi = self.plugin_configuration.get_integer_property_with_default(
            "ul_multi",
            1,
            valid_value_fn=self.__validate_minimum,
        )
        self.__ol_single = self.plugin_configuration.get_integer_property_with_default(
            "ol_single",
            1,
            valid_value_fn=self.__validate_minimum,
        )
        self.__ol_multi = self.plugin_configuration.get_integer_property_with_default(
            "ol_multi",
            1,
            valid_value_fn=self.__validate_minimum,
        )

    def query_config(self) -> List[QueryConfigItem]:
        """
        Query to find out the configuration that the rule is using.
        """
        return [
            QueryConfigItem("ul_single", self.__ul_single),
            QueryConfigItem("ul_multi", self.__ul_multi),
            QueryConfigItem("ol_single", self.__ol_single),
            QueryConfigItem("ol_multi", self.__ol_multi),
        ]

    def starting_new_file(self) -> None:
        """
        Event that the a new file to be scanned is starting.
        """
        self.__list_stack = []
        self.__list_tokens = []
        self.__current_list_parent = None
        self.__paragraph_count_map = {}
        self.__frank.starting_new_file()

    def __handle_list_end(self, context: PluginScanContext) -> None:
        for list_token in self.__list_tokens[-1]:
            this_list_token_paragraph_count = 0
            if str(list_token) in self.__paragraph_count_map:
                this_list_token_paragraph_count = self.__paragraph_count_map[
                    str(list_token)
                ]
            if self.__list_tokens[-1][0].is_ordered_list_start:
                required_spaces = (
                    self.__ol_multi
                    if this_list_token_paragraph_count > 1
                    else self.__ol_single
                )
            else:
                required_spaces = (
                    self.__ul_multi
                    if this_list_token_paragraph_count > 1
                    else self.__ul_single
                )
            # if self.__debug:
            #     print(">>" + str(list_token))
            #     print(
            #         ">>"
            #         + str(this_list_token_paragraph_count)
            #         + "..."
            #         + str(required_spaces)
            #     )
            #     print(">>" + str(self.__paragraph_count_map))
            if str(list_token) in self.__paragraph_count_map:
                del self.__paragraph_count_map[str(list_token)]
            # if self.__debug:
            #     print(">>" + str(self.__paragraph_count_map))

            list_start_token = cast(ListStartMarkdownToken, list_token)
            delta = list_start_token.indent_level - list_token.column_number
            if self.__list_stack[-1].is_ordered_list_start:
                delta -= len(list_start_token.list_start_content)
            # if self.__debug:
            #     print("y=" + str(list_token).replace(ParserHelper.newline_character, "\\n"))
            #     print(
            #         "token_index="
            #         + str(token_index)
            #         + "::"
            #         + str(list_token.indent_level)
            #         + ":"
            #         + str(list_token.column_number)
            #     )
            #     print(
            #         "delta=" + str(delta) + ",required_spaces=" + str(required_spaces)
            #     )
            if delta != required_spaces:
                extra_error_information = (
                    f"Expected: {required_spaces}; Actual: {delta}"
                )
                self.__report_or_fix(
                    context,
                    list_start_token,
                    extra_error_information,
                    delta - required_spaces,
                )

    def __report_or_fix(
        self,
        context: PluginScanContext,
        token: ListStartMarkdownToken,
        extra_error_information: str,
        adjust_amount: int,
    ) -> None:
        if context.in_fix_mode:
            self.register_fix_token_request(
                context,
                token,
                "next_token",
                "indent_level",
                token.indent_level - adjust_amount,
            )
            self.__frank.register(token, adjust_amount)
        else:
            self.report_next_token_error(
                context, token, extra_error_information=extra_error_information
            )

    def __next_token_list_start(self, token: MarkdownToken) -> None:
        self.__list_stack.append(token)
        self.__list_tokens.append([])
        self.__list_tokens[-1].append(token)
        self.__current_list_parent = token
        self.__frank.list_start(token)

    def __next_token_list_end_registrations(
        self,
        context: PluginScanContext,
        token: MarkdownToken,
        registration_map: Dict[MarkdownToken, int],
    ) -> None:
        end_token = cast(EndMarkdownToken, token)
        list_token = cast(ListStartMarkdownToken, end_token.start_markdown_token)
        if list_token.leading_spaces:
            split_leading_spaces = list_token.leading_spaces.split("\n")
            for registered_token, adj in registration_map.items():
                start, stop = self.__frank.get_start_stop(registered_token)
                for next_index in range(start, stop):
                    if (
                        next_index == len(split_leading_spaces) - 1
                        or not split_leading_spaces[next_index]
                    ):
                        continue
                    split_leading_spaces[next_index] = (
                        split_leading_spaces[next_index][:-adj]
                        if adj > 0
                        else split_leading_spaces[next_index] + (" " * -adj)
                    )
            rebuilt_leading_spaces = "\n".join(split_leading_spaces)
            if rebuilt_leading_spaces != list_token.leading_spaces:
                self.register_fix_token_request(
                    context,
                    list_token,
                    "next_token",
                    "leading_spaces",
                    rebuilt_leading_spaces,
                )

    def __next_token_list_end(
        self, context: PluginScanContext, token: MarkdownToken
    ) -> None:
        self.__frank.list_end()
        self.__handle_list_end(context)
        if registration_map := self.__frank.get_registrations():
            self.__next_token_list_end_registrations(context, token, registration_map)

        self.__frank.list_end_cleanup()
        del self.__list_stack[-1]
        del self.__list_tokens[-1]
        self.__current_list_parent = None
        if self.__list_tokens:
            # if self.__debug:
            #     print("__list_stack-->" + str(self.__list_stack))
            #     print("__list_tokens-->" + str(self.__list_tokens))
            self.__current_list_parent = self.__list_tokens[-1][-1]

    def next_token(self, context: PluginScanContext, token: MarkdownToken) -> None:
        """
        Event that a new token is being processed.
        """
        # if self.__debug:
        #     print(">>>>" + str(token))
        #     print("parent-->" + str(self.__current_list_parent))
        #     print("paragraph_count_map-->" + str(self.__paragraph_count_map))
        if token.is_list_start:
            self.__next_token_list_start(token)
        elif token.is_list_end:
            self.__next_token_list_end(context, token)
        elif token.is_new_list_item:
            self.__list_tokens[-1].append(token)
            self.__current_list_parent = token
            self.__frank.new_list_item(token)
        elif token.is_paragraph and self.__current_list_parent:
            new_count = (
                self.__paragraph_count_map[str(self.__current_list_parent)]
                if str(self.__current_list_parent) in self.__paragraph_count_map
                else 0
            )
            self.__paragraph_count_map[str(self.__current_list_parent)] = new_count + 1
        # if self.__debug:
        #     print("parent-->" + str(self.__current_list_parent))

        self.__frank.next_token(token)


# pylint: enable=too-many-instance-attributes



"""
Module to provide for an error in fixing a document.
"""


class BadPluginFixError(Exception):
    """
    Class to provide for an error in fixing a document.
    """

    def __init__(self, formatted_message: str):
        super().__init__(formatted_message)



from typing import List, Optional, Tuple, cast

from pymarkdown.container_blocks.container_grab_bag import POGGER
from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.parser_state import ParserState
from pymarkdown.general.position_marker import PositionMarker
from pymarkdown.general.requeue_line_info import RequeueLineInfo
from pymarkdown.leaf_blocks.table_block_continuation_helper import (
    TableBlockContinuationHelper,
)
from pymarkdown.leaf_blocks.table_block_parse_helper import TableParseHelper
from pymarkdown.leaf_blocks.table_block_tuple import TableTuple
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.tokens.stack_token import TableBlockStackToken


class TableBlockHelper:

    @staticmethod
    def process_table_rows(
        parser_state: ParserState,
        position_marker: PositionMarker,
        remaining_line_to_parse: str,
        extracted_whitespace: str,
        unmodified_line_to_parse: str,
        original_stack_depth: int,
        original_document_depth: int,
        original_line: str,
    ) -> Tuple[bool, bool, bool, Optional[RequeueLineInfo], List[MarkdownToken]]:
        """
        Process a table row.  Note, this requires a lot of work to
        handle properly because it is not until a new line that we
        can determine if the table is correctly formed.
        """
        line_to_parse = position_marker.text_to_parse
        start_index: int = position_marker.index_number
        lines_to_requeue: List[str] = []

        POGGER.debug(">>remaining_line_to_parse>:$:<", remaining_line_to_parse)
        POGGER.debug(">>line_to_parse>:$:<", line_to_parse)
        POGGER.debug(">>start_index>:$:<", start_index)

        (
            line_to_parse,
            unmodified_line_to_parse,
            remaining_line_to_parse,
            start_index,
            is_blank_line,
        ) = TableBlockHelper.__handle_table_rows_init(
            remaining_line_to_parse,
            line_to_parse,
            start_index,
            original_line,
            unmodified_line_to_parse,
        )

        (
            was_started,
            table_stack_token,
            original_stack_depth,
            original_document_depth,
            line_to_parse,
            extracted_whitespace,
            start_index,
        ) = TableBlockHelper.__handle_table_rows_started(
            parser_state,
            original_stack_depth,
            original_document_depth,
            line_to_parse,
            extracted_whitespace,
            start_index,
        )
        POGGER.debug(">>line_to_parse>:$:<", line_to_parse)
        line_to_parse_size = len(line_to_parse)

        (
            did_complete_table,
            end_table_index,
            parsed_table_tuple,
            is_blank_line,
            line_to_parse,
        ) = TableBlockHelper.__handle_table_rows_processing(
            parser_state,
            line_to_parse,
            start_index,
            extracted_whitespace,
            is_blank_line,
            was_started,
            line_to_parse_size,
            remaining_line_to_parse,
            lines_to_requeue,
            unmodified_line_to_parse,
        )
        (
            did_pause_table,
            force_ignore_first_as_table,
            new_tokens,
        ) = TableBlockContinuationHelper.determine_continue_or_stop(
            parser_state,
            position_marker,
            was_started,
            remaining_line_to_parse,
            extracted_whitespace,
            unmodified_line_to_parse,
            original_stack_depth,
            original_document_depth,
            end_table_index,
            line_to_parse_size,
            is_blank_line,
            did_complete_table,
            parsed_table_tuple,
            lines_to_requeue,
        )
        if lines_to_requeue:
            assert (
                table_stack_token is not None
            ), "Stack token is created before determining requeue, therefore it must be declared."
            TableBlockHelper.__prepare_for_requeue(
                parser_state,
                table_stack_token,
                did_complete_table,
                original_stack_depth,
                original_document_depth,
                lines_to_requeue,
            )
            ## Different from LRD.
            # force_ignore_first_as_table = True  # TODO check this
            ## Different from LRD.
            requeue_line_info = RequeueLineInfo(
                lines_to_requeue, False, force_ignore_first_as_table
            )
            ## Different from LRD.
            parser_state.abc(requeue_line_info, table_stack_token)
            ## Different from LRD.
        else:
            requeue_line_info = None

        return (
            did_complete_table or end_table_index != -1,
            did_complete_table,
            did_pause_table,
            requeue_line_info,
            new_tokens,
        )

    @staticmethod
    def __handle_table_rows_init(
        remaining_line_to_parse: str,
        line_to_parse: str,
        start_index: int,
        original_line: str,
        unmodified_line_to_parse: str,
    ) -> Tuple[str, str, str, int, bool]:
        is_blank_line = not line_to_parse and not start_index

        POGGER.debug(">>original_line>:$:<", original_line)
        if ParserHelper.tab_character in original_line and not is_blank_line:
            POGGER.debug(">>tabified>:$:<", original_line)

            first_character_to_parse = line_to_parse[start_index]
            POGGER.debug(">>xx>:$:<", first_character_to_parse)
            first_character_to_parse_index = original_line.find(
                first_character_to_parse
            )
            assert (
                first_character_to_parse_index != -1
            ), "Character not found within string."

            line_to_parse = original_line[first_character_to_parse_index:]
            unmodified_line_to_parse = original_line
            remaining_line_to_parse = line_to_parse
            start_index = 0

        POGGER.debug(">>line_to_parse>:$:<", line_to_parse)
        return (
            line_to_parse,
            unmodified_line_to_parse,
            remaining_line_to_parse,
            start_index,
            is_blank_line,
        )

    @staticmethod
    def __handle_table_rows_started(
        parser_state: ParserState,
        original_stack_depth: int,
        original_document_depth: int,
        line_to_parse: str,
        extracted_whitespace: str,
        start_index: int,
    ) -> Tuple[bool, Optional[TableBlockStackToken], int, int, str, str, int]:
        table_stack_token: Optional[TableBlockStackToken] = None

        if was_started := parser_state.token_stack[-1].was_table_block_started:
            table_stack_token = cast(TableBlockStackToken, parser_state.token_stack[-1])
            assert (
                table_stack_token.original_stack_depth is not None
                and table_stack_token.original_document_depth is not None
            ), "stack and document depth must both be defined."
            original_stack_depth, original_document_depth = (
                table_stack_token.original_stack_depth,
                table_stack_token.original_document_depth,
            )
            POGGER.debug(
                ">>continuation_lines>>$<<",
                table_stack_token.continuation_lines,
            )
            line_to_parse = table_stack_token.add_joined_lines_before_suffix(
                line_to_parse
            )
            (
                start_index,
                extracted_whitespace,
            ) = ParserHelper.extract_ascii_whitespace_verified(line_to_parse, 0)

            POGGER.debug(">>line_to_parse>>$<<", line_to_parse)
        return (
            was_started,
            table_stack_token,
            original_stack_depth,
            original_document_depth,
            line_to_parse,
            extracted_whitespace,
            start_index,
        )

    @staticmethod
    def __handle_table_rows_processing(
        parser_state: ParserState,
        line_to_parse: str,
        start_index: int,
        extracted_whitespace: str,
        is_blank_line: bool,
        was_started: bool,
        line_to_parse_size: int,
        remaining_line_to_parse: str,
        lines_to_requeue: List[str],
        unmodified_line_to_parse: str,
    ) -> Tuple[bool, int, Optional[TableTuple], bool, str]:
        if was_started:
            POGGER.debug(">>parse_table>>was_started")

            (
                did_complete_table,
                end_table_index,
                parsed_table_tuple,
            ) = TableParseHelper.parse_table(
                parser_state,
                line_to_parse,
                start_index,
                extracted_whitespace,
                is_blank_line,
                remaining_line_to_parse,
                was_started,
            )
            POGGER.debug(
                ">>parse_table>>was_started>>did_complete_table>>$"
                + ">>end_table_index>>$>>len(line_to_parse)>>$",
                did_complete_table,
                end_table_index,
                line_to_parse_size,
            )

            if not (
                did_complete_table
                or (
                    not is_blank_line
                    and not did_complete_table
                    and (end_table_index == line_to_parse_size)
                )
            ):
                POGGER.debug(">>parse_table>>was_started>>GOT HARD FAILURE")
                (
                    is_blank_line,
                    line_to_parse,
                    did_complete_table,
                    end_table_index,
                    parsed_table_tuple,
                ) = TableBlockHelper.__process_table_hard_failure(
                    parser_state,
                    remaining_line_to_parse,
                    lines_to_requeue,
                    unmodified_line_to_parse,
                    was_started,
                )
        else:
            (
                did_complete_table,
                end_table_index,
                parsed_table_tuple,
            ) = TableParseHelper.parse_table(
                parser_state,
                line_to_parse,
                start_index,
                extracted_whitespace,
                is_blank_line,
                remaining_line_to_parse,
                was_started,
            )
            POGGER.debug(
                ">>parse_table>>did_complete_table>>$>>end_table_index>>$>>len(line_to_parse)>>$",
                did_complete_table,
                end_table_index,
                line_to_parse_size,
            )
        return (
            did_complete_table,
            end_table_index,
            parsed_table_tuple,
            is_blank_line,
            line_to_parse,
        )

    @staticmethod
    def __prepare_for_requeue_reset_markdown_token(
        parser_state: ParserState, table_stack_token: TableBlockStackToken
    ) -> None:
        assert (
            table_stack_token.last_block_quote_markdown_token_index is not None
        ), "Index must be defined by now."
        POGGER.debug(
            ">>XXXXXX>>last_block_quote_markdown_token_index:$:",
            table_stack_token.last_block_quote_markdown_token_index,
        )
        POGGER.debug(
            ">>XXXXXX>>st-now:$:",
            parser_state.token_document[
                table_stack_token.last_block_quote_markdown_token_index
            ],
        )

        del parser_state.token_document[
            table_stack_token.last_block_quote_markdown_token_index
        ]
        assert (
            table_stack_token.copy_of_last_block_quote_markdown_token is not None
        ), "Token must be defined by now."
        parser_state.token_document.insert(
            table_stack_token.last_block_quote_markdown_token_index,
            table_stack_token.copy_of_last_block_quote_markdown_token,
        )
        assert (
            table_stack_token.last_block_quote_stack_token is not None
        ), "Token must be defined by now."
        POGGER.debug(
            "last_block_quote_stack_token>:$:<",
            table_stack_token.last_block_quote_stack_token,
        )
        table_stack_token.last_block_quote_stack_token.reset_matching_markdown_token(
            table_stack_token.copy_of_last_block_quote_markdown_token
        )
        POGGER.debug(
            "last_block_quote_stack_token>:$:<",
            table_stack_token.last_block_quote_stack_token,
        )

    @staticmethod
    def __prepare_for_requeue_reset_document_and_stack(
        parser_state: ParserState,
        table_stack_token: TableBlockStackToken,
        original_stack_depth: int,
        original_document_depth: int,
    ) -> None:
        POGGER.debug(">>XXXXXX>>original_stack_depth:$:", original_stack_depth)
        POGGER.debug(">>XXXXXX>>token_stack_depth:$:", len(parser_state.token_stack))
        POGGER.debug(">>XXXXXX>>token_stack(before):$:", parser_state.token_stack)
        POGGER.debug(
            ">>XXXXXX>>copy_of_token_stack:$:", parser_state.copy_of_token_stack
        )
        POGGER.debug(
            ">>table_stack_token>>copy_of_token_stack:$:",
            table_stack_token.copy_of_token_stack,
        )
        if len(parser_state.token_stack) >= original_stack_depth:
            while (
                len(parser_state.token_stack) > original_stack_depth
                ## Different from LRD.
                and not parser_state.token_stack[-1].is_block_quote
                ## Different from LRD.
            ):
                del parser_state.token_stack[-1]
        else:
            while len(parser_state.token_stack):
                del parser_state.token_stack[-1]
            assert (
                table_stack_token.copy_of_token_stack is not None
            ), "Token must be defined by now."
            parser_state.token_stack.extend(table_stack_token.copy_of_token_stack)
        POGGER.debug(">>XXXXXX>>token_stack(after):$:", parser_state.token_stack)

        ## Different from LRD.
        POGGER.debug(">>XXXXXX>>original_document_depth:$:", original_document_depth)
        POGGER.debug(
            ">>XXXXXX>>token_document_depth:$:",
            len(parser_state.token_document),
        )
        POGGER.debug(">>XXXXXX>>token_document(before):$:", parser_state.token_document)
        while (
            len(parser_state.token_document) > original_document_depth
            and not parser_state.token_document[-1].is_block_quote_start
        ):
            del parser_state.token_document[-1]
        ## Different from LRD.
        POGGER.debug(">>XXXXXX>>token_document(after):$:", parser_state.token_document)

    @staticmethod
    def __prepare_for_requeue(
        parser_state: ParserState,
        table_stack_token: TableBlockStackToken,
        did_complete_table: bool,
        original_stack_depth: int,
        original_document_depth: int,
        lines_to_requeue: List[str],
    ) -> None:
        # This works because in most cases, we add things.  However, in cases like
        # an indented code block, we process the "is it indented enough" and close
        # that block before hitting this.  As such, we have a special case to take
        # care of that.  In the future, will possibly want to do something instead of
        # original_document_depth and stack, such as passing in a copy of the both
        # elements so they can be reset on the rewind.
        # i.e. icode would go back on stack, end-icode would not be in document.
        POGGER.debug("lines_to_requeue:$:", lines_to_requeue)
        POGGER.debug(
            ">>XXXXXX>>copy_of_last_block_quote_markdown_token:$:",
            table_stack_token.copy_of_last_block_quote_markdown_token,
        )
        if not did_complete_table:
            if table_stack_token.copy_of_last_block_quote_markdown_token:
                TableBlockHelper.__prepare_for_requeue_reset_markdown_token(
                    parser_state, table_stack_token
                )
            TableBlockHelper.__prepare_for_requeue_reset_document_and_stack(
                parser_state,
                table_stack_token,
                original_stack_depth,
                original_document_depth,
            )

    @staticmethod
    def __process_table_hard_failure(
        parser_state: ParserState,
        remaining_line_to_parse: str,
        lines_to_requeue: List[str],
        unmodified_line_to_parse: str,
        was_started: bool,
    ) -> Tuple[
        bool,
        str,
        bool,
        int,
        Optional[TableTuple],
    ]:
        """
        In cases of a hard failure, we have had continuations to the original line
        that make it a bit more difficult to figure out if we have an actual good
        LRD in the mix somehow.  So take lines off the end while we have lines.
        """
        (
            is_blank_line,
            line_to_parse,
            did_complete_table,
            end_table_index,
            parsed_table_tuple,
        ) = (None, None, None, None, None)
        table_stack_token = cast(TableBlockStackToken, parser_state.token_stack[-1])

        POGGER.debug(">>remaining_line_to_parse>>add>:$<<", remaining_line_to_parse)
        POGGER.debug(">>unmodified_line_to_parse>>add>:$<<", unmodified_line_to_parse)
        assert unmodified_line_to_parse.endswith(
            remaining_line_to_parse
        ), "Current line must end with the remaining text."
        ## Different from LRD.
        table_stack_token.add_continuation_line(remaining_line_to_parse)
        table_stack_token.add_unmodified_line(unmodified_line_to_parse)
        ## Different from LRD.

        ## Different from LRD.
        try_again = len(table_stack_token.continuation_lines) > 2
        if not try_again:
            is_blank_line = False
            line_to_parse = remaining_line_to_parse
            did_complete_table = False
            end_table_index = -1
        xdf = 2 if len(table_stack_token.continuation_lines) == 2 else 1
        while xdf:
            ## Different from LRD.

            lines_to_requeue.append(table_stack_token.unmodified_lines[-1])
            del table_stack_token.continuation_lines[-1]
            del table_stack_token.unmodified_lines[-1]

            ## Different from LRD.
            if try_again:
                (
                    is_blank_line,
                    line_to_parse,
                ) = True, table_stack_token.add_joined_lines_before_suffix("")
                line_to_parse = line_to_parse[:-1]
                start_index, extracted_whitespace = (
                    ParserHelper.extract_spaces_verified(line_to_parse, 0)
                )
                (
                    did_complete_table,
                    end_table_index,
                    parsed_table_tuple,
                ) = TableParseHelper.parse_table(
                    parser_state,
                    line_to_parse,
                    start_index,
                    extracted_whitespace,
                    is_blank_line,
                    None,
                    was_started,
                )
            xdf -= 1
        ## Different from LRD.
        assert is_blank_line is not None, "while loop must have executed at least once."
        assert line_to_parse is not None, "while loop must have executed at least once."
        assert (
            did_complete_table is not None
        ), "while loop must have executed at least once."
        assert (
            end_table_index is not None
        ), "while loop must have executed at least once."
        return (
            is_blank_line,
            line_to_parse,
            did_complete_table,
            end_table_index,
            parsed_table_tuple,
        )

    @staticmethod
    def handle_table_leaf_block(
        parser_state: ParserState,
        outer_processed: bool,
        position_marker: PositionMarker,
        leaf_token_whitespace: str,
        remaining_line_to_parse: str,
        ignore_table_start: bool,
        pre_tokens: List[MarkdownToken],
        original_line: str,
        requeue_line_info: Optional[RequeueLineInfo],
    ) -> Tuple[bool, Optional[RequeueLineInfo]]:
        """
        TBD

        NOTE: Since `TableBlockHelper.handle_table_leaf_block` is called
        from `ContainerBlockLeafProcessor.handle_leaf_block` after
        `LinkReferenceDefinitionHelper.handle_link_reference_definition_leaf_block`,
        need to ensure that we handle the requeue_line_info properly.
        """
        POGGER.debug(
            "handle_table_leaf_block>>pre_tokens>>$<<",
            pre_tokens,
        )

        ## Different from LRD.
        if (
            parser_state.parse_properties.is_tables_enabled
            and not outer_processed
            and not ignore_table_start
        ):
            ## Different from LRD.
            POGGER.debug(
                "handle_table_leaf_block>>outer_processed>>$",
                position_marker.text_to_parse[position_marker.index_number :],
            )
            assert (
                parser_state.original_line_to_parse is not None
            ), "Original line must be defined by now."
            (
                outer_processed,
                _,  # did_complete_table,
                _,  # did_pause_table,
                requeue_line_info,
                new_tokens,
            ) = TableBlockHelper.process_table_rows(
                parser_state,
                position_marker,
                remaining_line_to_parse,
                leaf_token_whitespace,
                parser_state.original_line_to_parse,
                parser_state.original_stack_depth,
                parser_state.original_document_depth,
                original_line,
            )
            if requeue_line_info:
                outer_processed = True
                POGGER.debug(
                    "handle_table_leaf_block>>outer_processed>>$<lines_to_requeue<$<$",
                    outer_processed,
                    requeue_line_info.lines_to_requeue,
                    len(requeue_line_info.lines_to_requeue),
                )
            else:
                POGGER.debug(
                    "handle_table_leaf_block>>outer_processed>>$<lines_to_requeue<(None)",
                    outer_processed,
                )
        else:
            ## Different from LRD.
            new_tokens = []
        ## Different from LRD.

        POGGER.debug("handle_table_leaf_block>>pre_tokens>>$<<", pre_tokens)
        pre_tokens.extend(new_tokens)
        POGGER.debug("handle_table_leaf_block>>pre_tokens>>$<<", pre_tokens)
        return outer_processed, requeue_line_info



"""
Module to provide for an encapsulation of the inline emphasis element.
"""

from typing import Optional, cast

from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.tokens.inline_markdown_token import InlineMarkdownToken
from pymarkdown.tokens.markdown_token import EndMarkdownToken, MarkdownToken
from pymarkdown.transform_gfm.transform_state import TransformState
from pymarkdown.transform_markdown.markdown_transform_context import (
    MarkdownTransformContext,
    RegisterHtmlTransformHandlersProtocol,
    RegisterMarkdownTransformHandlersProtocol,
)


class EmphasisMarkdownToken(InlineMarkdownToken):
    """
    Class to provide for an encapsulation of the inline emphasis element.
    """

    def __init__(
        self,
        emphasis_length: int,
        emphasis_character: str,
        line_number: int = 0,
        column_number: int = 0,
    ) -> None:
        self.__emphasis_length, self.__emphasis_character = (
            emphasis_length,
            emphasis_character,
        )
        InlineMarkdownToken.__init__(
            self,
            MarkdownToken._token_inline_emphasis,
            MarkdownToken.extra_data_separator.join(
                [str(emphasis_length), emphasis_character]
            ),
            line_number=line_number,
            column_number=column_number,
            requires_end_token=True,
            can_force_close=False,
        )

    # pylint: disable=protected-access
    @staticmethod
    def get_markdown_token_type() -> str:
        """
        Get the type of markdown token for rehydration purposes.
        """
        return MarkdownToken._token_inline_emphasis

    # pylint: enable=protected-access

    @property
    def emphasis_length(self) -> int:
        """
        Returns the length of the current emphasis text.
        """
        return self.__emphasis_length

    @property
    def emphasis_character(self) -> str:
        """
        Returns the character used for the current emphasis text.
        """
        return self.__emphasis_character

    def register_for_markdown_transform(
        self, registration_function: RegisterMarkdownTransformHandlersProtocol
    ) -> None:
        """
        Register any rehydration handlers for leaf markdown tokens.
        """
        registration_function(
            EmphasisMarkdownToken,
            EmphasisMarkdownToken.__rehydrate_inline_emphaisis,
            EmphasisMarkdownToken.__rehydrate_inline_emphaisis_end,
        )

    @staticmethod
    def __rehydrate_inline_emphaisis(
        context: MarkdownTransformContext,
        current_token: MarkdownToken,
        previous_token: Optional[MarkdownToken],
    ) -> str:
        """
        Rehydrate the emphasis text from the token.
        """
        _ = previous_token

        emphasis_token = cast(EmphasisMarkdownToken, current_token)
        return (
            ""
            if context.block_stack[-1].is_inline_link
            else ParserHelper.repeat_string(
                emphasis_token.emphasis_character, emphasis_token.emphasis_length
            )
        )

    @staticmethod
    def __rehydrate_inline_emphaisis_end(
        context: MarkdownTransformContext,
        current_token: MarkdownToken,
        previous_token: Optional[MarkdownToken],
        next_token: Optional[MarkdownToken],
    ) -> str:
        """
        Rehydrate the emphasis end text from the token.
        """
        _ = (previous_token, next_token)

        emphasis_end_token = cast(EndMarkdownToken, current_token)
        emphasis_token = cast(
            EmphasisMarkdownToken, emphasis_end_token.start_markdown_token
        )

        return (
            ""
            if context.block_stack[-1].is_inline_link
            else ParserHelper.repeat_string(
                emphasis_token.emphasis_character,
                emphasis_token.emphasis_length,
            )
        )

    @staticmethod
    def register_for_html_transform(
        register_handlers: RegisterHtmlTransformHandlersProtocol,
    ) -> None:
        """
        Register any functions required to generate HTML from the tokens.
        """
        register_handlers(
            EmphasisMarkdownToken,
            EmphasisMarkdownToken.__handle_start_emphasis_token,
            EmphasisMarkdownToken.__handle_end_emphasis_token,
        )

    @staticmethod
    def __handle_start_emphasis_token(
        output_html: str,
        next_token: MarkdownToken,
        transform_state: TransformState,
    ) -> str:
        _ = transform_state

        emphasis_token = cast(EmphasisMarkdownToken, next_token)
        if emphasis_token.emphasis_character == "~":
            return f"{output_html}<del>"
        return "".join(
            [
                output_html,
                "<em>" if emphasis_token.emphasis_length == 1 else "<strong>",
            ]
        )

    @staticmethod
    def __handle_end_emphasis_token(
        output_html: str,
        next_token: MarkdownToken,
        transform_state: TransformState,
    ) -> str:
        _ = transform_state

        end_token = cast(EndMarkdownToken, next_token)
        emphasis_token = cast(EmphasisMarkdownToken, end_token.start_markdown_token)

        if emphasis_token.emphasis_character == "~":
            return f"{output_html}</del>"
        return "".join(
            [
                output_html,
                "</em>" if emphasis_token.emphasis_length == 1 else "</strong>",
            ]
        )



"""
Module to provide for special tokens that represent exceptional inline elements.
"""

from typing import Optional

from pymarkdown.tokens.text_markdown_token import TextMarkdownToken


class SpecialTextMarkdownToken(TextMarkdownToken):
    """
    Class to provide for special tokens that represent exceptional inline elements.
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        token_text: str,
        repeat_count: int,
        preceding_two: Optional[str],
        following_two: Optional[str],
        is_active: bool = True,
        line_number: int = 0,
        column_number: int = 0,
    ):
        (
            self.__repeat_count,
            self.__is_active,
            self.__preceding_two,
            self.__following_two,
        ) = (repeat_count, is_active, preceding_two, following_two)
        TextMarkdownToken.__init__(
            self,
            token_text,
            "",
            "",
            line_number=line_number,
            column_number=column_number,
            is_special=True,
        )

    # pylint: enable=too-many-arguments

    @property
    def is_active(self) -> bool:
        """
        Returns a value indicating whether this special text is still active.
        """
        return self.__is_active

    @property
    def repeat_count(self) -> int:
        """
        Returns the repeat count for the special text element.
        """
        return self.__repeat_count

    @property
    def preceding_two(self) -> Optional[str]:
        """
        Returns the preceding two characters before this token.
        """
        return self.__preceding_two

    @property
    def following_two(self) -> Optional[str]:
        """
        Returns the following two characters before this token.
        """
        return self.__following_two

    def deactivate(self) -> None:
        """
        Mark this special token as deactivated.
        """
        self.__is_active = False

    def adjust_token_text_by_repeat_count(self) -> None:
        """
        Adjust the token's text by the repeat count.
        """
        self._set_token_text(self.token_text[: self.repeat_count])

    def reduce_repeat_count(
        self, emphasis_length: int, adjust_column_number: bool = False
    ) -> None:
        """
        Reduce the repeat count by the specified amount.
        """
        self.__repeat_count -= emphasis_length
        if adjust_column_number:
            self._set_column_number(self.column_number + emphasis_length)

    def show_process_emphasis(self) -> str:
        """
        Independent of the __str__ function, provide extra information.
        """
        return (
            f">>active={self.__is_active},repeat={self.__repeat_count},preceding='{self.__preceding_two}',"
            + f"following='{self.__following_two}':{self}"
        )



"""
Module to provide for an encapsulation of the thematic break element.
"""

from typing import Optional, Union, cast

from typing_extensions import override

from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.position_marker import PositionMarker
from pymarkdown.tokens.leaf_markdown_token import LeafMarkdownToken
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.transform_gfm.transform_state import TransformState
from pymarkdown.transform_markdown.markdown_transform_context import (
    MarkdownTransformContext,
    RegisterHtmlTransformHandlersProtocol,
    RegisterMarkdownTransformHandlersProtocol,
)


class ThematicBreakMarkdownToken(LeafMarkdownToken):
    """
    Class to provide for an encapsulation of the thematic break element.
    """

    def __init__(
        self,
        start_character: str,
        extracted_whitespace: str,
        rest_of_line: str,
        position_marker: PositionMarker,
    ) -> None:
        self.__start_character = start_character
        self.__extracted_whitespace = extracted_whitespace
        self.__rest_of_line = rest_of_line
        LeafMarkdownToken.__init__(
            self,
            MarkdownToken._token_thematic_break,
            self.__compose_extra_data_field(),
            position_marker=position_marker,
            extracted_whitespace=extracted_whitespace,
        )

    # pylint: disable=protected-access
    @staticmethod
    def get_markdown_token_type() -> str:
        """
        Get the type of markdown token for rehydration purposes.
        """
        return MarkdownToken._token_thematic_break

    # pylint: enable=protected-access

    @property
    def rest_of_line(self) -> str:
        """
        Returns any whitespace that was extracted before the processing of this element occurred.
        """
        return self.__rest_of_line

    def __compose_extra_data_field(self) -> str:
        """
        Compose the object's self.extra_data field from the local object's variables.
        """
        x = MarkdownToken.extra_data_separator.join(
            [self.__start_character, self.__extracted_whitespace, self.__rest_of_line]
        )
        self._set_extra_data(x)
        return x

    @override
    def _modify_token(self, field_name: str, field_value: Union[str, int]) -> bool:
        if field_name == "start_character" and isinstance(field_value, str):
            self.__start_character = field_value
            self.__compose_extra_data_field()
            return True
        if field_name == "rest_of_line" and isinstance(field_value, str):
            self.__rest_of_line = field_value
            self.__compose_extra_data_field()
            return True
        return super()._modify_token(field_name, field_value)

    def register_for_markdown_transform(
        self, registration_function: RegisterMarkdownTransformHandlersProtocol
    ) -> None:
        """
        Register any rehydration handlers for leaf markdown tokens.
        """
        registration_function(
            ThematicBreakMarkdownToken,
            ThematicBreakMarkdownToken.__rehydrate_thematic_break,
            None,
        )

    @staticmethod
    def __rehydrate_thematic_break(
        context: MarkdownTransformContext,
        current_token: MarkdownToken,
        previous_token: Optional[MarkdownToken],
    ) -> str:
        """
        Rehydrate the thematic break text from the token.
        """
        _ = previous_token, context

        current_thematic_token = cast(ThematicBreakMarkdownToken, current_token)
        return "".join(
            [
                current_thematic_token.extracted_whitespace,
                current_thematic_token.rest_of_line,
                ParserHelper.newline_character,
            ]
        )

    @staticmethod
    def register_for_html_transform(
        register_handlers: RegisterHtmlTransformHandlersProtocol,
    ) -> None:
        """
        Register any functions required to generate HTML from the tokens.
        """
        register_handlers(
            ThematicBreakMarkdownToken,
            ThematicBreakMarkdownToken.__handle_thematic_break_token,
            None,
        )

    @staticmethod
    def __handle_thematic_break_token(
        output_html: str,
        next_token: MarkdownToken,
        transform_state: TransformState,
    ) -> str:
        _ = (next_token, transform_state)

        token_parts = [output_html]
        if output_html and output_html[-1] != ParserHelper.newline_character:
            token_parts.append(ParserHelper.newline_character)
        token_parts.extend(["<hr />", ParserHelper.newline_character])
        return "".join(token_parts)



"""
Module to provide for an encapsulation of the inline code span element.
"""

from typing import Optional, Union, cast

from typing_extensions import override

from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.tokens.inline_markdown_token import InlineMarkdownToken
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.tokens.paragraph_markdown_token import ParagraphMarkdownToken
from pymarkdown.transform_gfm.transform_state import TransformState
from pymarkdown.transform_markdown.markdown_transform_context import (
    MarkdownTransformContext,
    RegisterHtmlTransformHandlersProtocol,
    RegisterMarkdownTransformHandlersProtocol,
)


class InlineCodeSpanMarkdownToken(InlineMarkdownToken):
    """
    Class to provide for an encapsulation of the inline code span element.
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        span_text: str,
        extracted_start_backticks: str,
        leading_whitespace: str,
        trailing_whitespace: str,
        is_in_table: bool,
        line_number: int,
        column_number: int,
    ) -> None:
        (
            self.__span_text,
            self.__extracted_start_backticks,
            self.__leading_whitespace,
            self.__trailing_whitespace,
            self.is_in_table,
        ) = (
            span_text,
            extracted_start_backticks,
            leading_whitespace,
            trailing_whitespace,
            is_in_table,
        )
        InlineMarkdownToken.__init__(
            self,
            MarkdownToken._token_inline_code_span,
            self.__compose_extra_data_field(),
            line_number=line_number,
            column_number=column_number,
        )

    def __compose_extra_data_field(self) -> str:
        """
        Compose the object's self.extra_data field from the local object's variables.
        """
        return MarkdownToken.extra_data_separator.join(
            [
                self.__span_text,
                self.__extracted_start_backticks,
                self.__leading_whitespace,
                self.__trailing_whitespace,
            ]
        )

    @override
    def _modify_token(self, field_name: str, field_value: Union[str, int]) -> bool:
        if field_name == "span_text" and isinstance(field_value, str):
            self.__span_text = field_value
            self._set_extra_data(self.__compose_extra_data_field())
            return True
        return False

    # pylint: disable=protected-access
    @staticmethod
    def get_markdown_token_type() -> str:
        """
        Get the type of markdown token for rehydration purposes.
        """
        return MarkdownToken._token_inline_code_span

    # pylint: enable=protected-access

    # pylint: enable=too-many-arguments
    @property
    def span_text(self) -> str:
        """
        Returns the text that is within the span.
        """
        return self.__span_text

    @property
    def extracted_start_backticks(self) -> str:
        """
        Returns the backticks that started the code span.
        """
        return self.__extracted_start_backticks

    @property
    def leading_whitespace(self) -> str:
        """
        Returns the whitespace at the start of the code span.
        """
        return self.__leading_whitespace

    @property
    def trailing_whitespace(self) -> str:
        """
        Returns the whitespace at the end of the code span.
        """
        return self.__trailing_whitespace

    def register_for_markdown_transform(
        self, registration_function: RegisterMarkdownTransformHandlersProtocol
    ) -> None:
        """
        Register any rehydration handlers for leaf markdown tokens.
        """
        registration_function(
            InlineCodeSpanMarkdownToken,
            InlineCodeSpanMarkdownToken.__rehydrate_inline_code_span,
            None,
        )

    @staticmethod
    def __rehydrate_inline_code_span(
        context: MarkdownTransformContext,
        current_token: MarkdownToken,
        previous_token: Optional[MarkdownToken],
    ) -> str:
        """
        Rehydrate the code span data from the token.
        """
        _ = previous_token

        if context.block_stack[-1].is_inline_link:
            return ""

        current_inline_token = cast(InlineCodeSpanMarkdownToken, current_token)
        span_text = ParserHelper.remove_all_from_text(
            current_inline_token.span_text, include_noops=True
        )
        leading_whitespace = ParserHelper.remove_all_from_text(
            current_inline_token.leading_whitespace
        )
        trailing_whitespace = ParserHelper.remove_all_from_text(
            current_inline_token.trailing_whitespace
        )

        if context.block_stack[-1].is_paragraph:
            block_paragraph_token = cast(
                ParagraphMarkdownToken, context.block_stack[-1]
            )
            (
                leading_whitespace,
                block_paragraph_token.rehydrate_index,
            ) = ParserHelper.recombine_string_with_whitespace(
                leading_whitespace,
                block_paragraph_token.extracted_whitespace,
                block_paragraph_token.rehydrate_index,
            )
            (
                span_text,
                block_paragraph_token.rehydrate_index,
            ) = ParserHelper.recombine_string_with_whitespace(
                span_text,
                block_paragraph_token.extracted_whitespace,
                block_paragraph_token.rehydrate_index,
            )
            (
                trailing_whitespace,
                block_paragraph_token.rehydrate_index,
            ) = ParserHelper.recombine_string_with_whitespace(
                trailing_whitespace,
                block_paragraph_token.extracted_whitespace,
                block_paragraph_token.rehydrate_index,
            )

        return "".join(
            [
                current_inline_token.extracted_start_backticks,
                leading_whitespace,
                span_text,
                trailing_whitespace,
                current_inline_token.extracted_start_backticks,
            ]
        )

    @staticmethod
    def register_for_html_transform(
        register_handlers: RegisterHtmlTransformHandlersProtocol,
    ) -> None:
        """
        Register any functions required to generate HTML from the tokens.
        """
        register_handlers(
            InlineCodeSpanMarkdownToken,
            InlineCodeSpanMarkdownToken.__handle_inline_code_span_token,
            None,
        )

    @staticmethod
    def __handle_inline_code_span_token(
        output_html: str,
        next_token: MarkdownToken,
        transform_state: TransformState,
    ) -> str:
        _ = transform_state

        code_span_token = cast(InlineCodeSpanMarkdownToken, next_token)
        span_text = code_span_token.span_text
        if code_span_token.is_in_table and "\\" in span_text and "|" in span_text:
            new_span_text = ""
            span_index = 0
            while span_index < len(span_text):
                next_bar_index, before_bar_text = ParserHelper.collect_until_character(
                    span_text, span_index, "|"
                )
                assert next_bar_index is not None
                assert before_bar_text is not None
                if next_bar_index < len(span_text):
                    assert (
                        before_bar_text and before_bar_text[-1] == "\\"
                    )  # for it to be in a code span within a table, it has to be escaped
                    before_bar_text = before_bar_text[:-1]
                    new_span_text = f"{new_span_text}{before_bar_text}|"
                else:
                    new_span_text = f"{new_span_text}{before_bar_text}"
                span_index = next_bar_index + 1
            span_text = new_span_text

        return "".join(
            [
                output_html,
                "<code>",
                ParserHelper.resolve_all_from_text(span_text),
                "</code>",
            ]
        )



"""
Module to handle the calculationof list looseness for the GRM transformer.
"""

import logging
from typing import List, Optional, Tuple, cast

from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.tokens.list_start_markdown_token import ListStartMarkdownToken
from pymarkdown.tokens.markdown_token import EndMarkdownToken, MarkdownToken

POGGER = ParserLogger(logging.getLogger(__name__))


class TransformToGfmListLooseness:
    """
    Class to handle the calculationof list looseness for the GRM transformer.
    """

    @staticmethod
    def calculate_list_looseness(
        actual_tokens: List[MarkdownToken],
        actual_token_index: int,
        next_token: ListStartMarkdownToken,
    ) -> bool:
        """
        Based on the first token in a list, compute the "looseness" of the list.
        """

        POGGER.debug("\n\n__calculate_list_looseness>>$", actual_token_index)
        is_loose, current_token_index, stack_count = False, actual_token_index + 1, 0
        while True:
            current_token = actual_tokens[current_token_index]
            (
                check_me,
                stack_count,
                stop_me,
                is_loose,
            ) = TransformToGfmListLooseness.__calculate_list_looseness_for_containers(
                current_token,
                stack_count,
                is_loose,
                False,
                actual_tokens,
                current_token_index,
            )
            if check_me:
                POGGER.debug("check-->?")
                is_loose = stop_me = TransformToGfmListLooseness.__is_token_loose(
                    actual_tokens, current_token_index
                )
                POGGER.debug("check-->Loose=$", is_loose)
            if stop_me:
                break
            current_token_index += 1

        assert current_token_index != len(
            actual_tokens
        ), "Index must be within the string."
        POGGER.debug(
            "__calculate_list_looseness<<$<<$\n\n",
            actual_token_index,
            is_loose,
        )
        next_token.is_loose = is_loose
        return is_loose

    # pylint: disable=too-many-arguments
    @staticmethod
    def __calculate_list_looseness_for_containers(
        current_token: MarkdownToken,
        stack_count: int,
        is_loose: bool,
        stop_me: bool,
        actual_tokens: List[MarkdownToken],
        current_token_index: int,
    ) -> Tuple[bool, int, bool, bool]:
        check_me = False
        if current_token.is_list_start:
            POGGER.debug("cll>>start list>>$", current_token)
            check_me, stack_count = TransformToGfmListLooseness.__handle_list_start(
                stack_count
            )
        elif current_token.is_new_list_item:
            POGGER.debug("cll>>new list item>>$", current_token)
            check_me = TransformToGfmListLooseness.__handle_new_list_item(
                current_token, stack_count
            )
        elif current_token.is_block_quote_start:
            POGGER.debug("cll>>start block quote>>$", current_token)
            stack_count = TransformToGfmListLooseness.__handle_block_quote_start(
                stack_count
            )
            check_me = True
        elif current_token.is_block_quote_end:
            POGGER.debug("cll>>end block quote>>$", current_token)
            (
                stop_me,
                is_loose,
                stack_count,
            ) = TransformToGfmListLooseness.__handle_block_quote_end(
                stack_count, actual_tokens, current_token_index, stop_me, is_loose
            )
        elif current_token.is_list_end:
            POGGER.debug("cll>>end list>>$", current_token)
            (
                stop_me,
                is_loose,
                stack_count,
            ) = TransformToGfmListLooseness.__handle_list_end(
                stack_count,
                is_loose,
                actual_tokens,
                current_token_index,
            )
            POGGER.debug("cll>>stop_me>>$", stop_me)
            POGGER.debug("cll>>is_loose>>$", is_loose)
        elif actual_tokens[current_token_index - 1].is_blank_line:
            POGGER.debug("cll>>__handle_blank_line>>")
            check_me = TransformToGfmListLooseness.__handle_blank_line(
                current_token, stack_count, actual_tokens, current_token_index
            )

        POGGER.debug(
            ">>stack_count>>$>>#$:$>>check=$",
            stack_count,
            current_token_index,
            actual_tokens[current_token_index],
            check_me,
        )
        return check_me, stack_count, stop_me, is_loose

    # pylint: enable=too-many-arguments

    @staticmethod
    def __handle_list_start(stack_count: int) -> Tuple[bool, int]:
        POGGER.debug(">>list--new>>$", stack_count)
        return stack_count == 0, stack_count + 1

    @staticmethod
    def __handle_new_list_item(current_token: MarkdownToken, stack_count: int) -> bool:
        assert not current_token.is_block, "Cannot be a block."
        POGGER.debug(">>list--item>>$", stack_count)
        return stack_count == 0

    @staticmethod
    def __handle_list_end(
        stack_count: int,
        is_loose: bool,
        actual_tokens: List[MarkdownToken],
        current_token_index: int,
    ) -> Tuple[bool, bool, int]:
        POGGER.debug(">>list--end>>$", stack_count)
        stop_me = not stack_count
        if not stop_me:
            stack_count -= 1
            if not stack_count:
                POGGER.debug("<<check!!")

                check_index = current_token_index + 1
                assert check_index < len(actual_tokens), "Index must be within list."
                POGGER.debug_with_visible_whitespace(
                    "<<check>>$", actual_tokens[check_index]
                )
                if not actual_tokens[check_index].is_list_end:
                    POGGER.debug("<END OF possibly multiple list ends")
                    search_back_index = current_token_index - 1
                    # while (
                    #     search_back_index >= 0
                    #     and actual_tokens[check_index].is_list_end
                    # ):
                    #     search_back_index -= 1
                    stop_me = TransformToGfmListLooseness.__is_token_loose(
                        actual_tokens, search_back_index + 1
                    )
                    is_loose = stop_me
                    if stop_me:
                        POGGER.debug("!!!latent-LOOSE!!!")
        POGGER.debug("<<list--end>>$", stack_count)
        return stop_me, is_loose, stack_count

    @staticmethod
    def __handle_block_quote_start(stack_count: int) -> int:
        POGGER.debug(">>block--new>>$", stack_count)
        return stack_count + 1

    @staticmethod
    def __handle_block_quote_end_calc(
        actual_tokens: List[MarkdownToken], search_index: int, current_token_index: int
    ) -> bool:
        stop_me = TransformToGfmListLooseness.__is_token_loose(
            actual_tokens, search_index + 1, True
        )
        if stop_me:
            blank_token = actual_tokens[search_index]
            search_index = current_token_index
            next_token = actual_tokens[search_index]
            while search_index >= 0 and not next_token.is_block_quote_start:
                search_index -= 1
                next_token = actual_tokens[search_index]
            block_quote_spaces_count = len(next_token.bleading_spaces.split("\n"))  # type: ignore
            end_block_quote_line = next_token.line_number + block_quote_spaces_count
            stop_me = not blank_token.line_number < end_block_quote_line
        return stop_me

    @staticmethod
    def __handle_block_quote_end(
        stack_count: int,
        actual_tokens: List[MarkdownToken],
        current_token_index: int,
        stop_me: bool,
        is_loose: bool,
    ) -> Tuple[bool, bool, int]:
        POGGER.debug(">>block--end>>$", stack_count)
        old_stack_count = stack_count
        stack_count -= 1
        if old_stack_count and not stack_count:
            search_index = current_token_index - 1
            found_end_token: Optional[MarkdownToken] = actual_tokens[search_index]
            while found_end_token is not None:
                found_end_token = None
                next_token = actual_tokens[search_index]
                if next_token.is_end_token and (
                    next_token.is_block_quote_end or next_token.is_list_end
                ):
                    found_end_token = next_token
                    search_index -= 1

            new_index = current_token_index + 1
            keep_checking = (
                new_index < len(actual_tokens) and actual_tokens[new_index].is_end_token
            ) and actual_tokens[new_index].is_list_end
            if not keep_checking:
                (stop_me) = TransformToGfmListLooseness.__handle_block_quote_end_calc(
                    actual_tokens, search_index, current_token_index
                )
                is_loose = stop_me
        return stop_me, is_loose, stack_count

    @staticmethod
    def __handle_blank_line(
        current_token: MarkdownToken,
        stack_count: int,
        actual_tokens: List[MarkdownToken],
        current_token_index: int,
    ) -> bool:
        search_back_index = current_token_index - 2
        while actual_tokens[search_back_index].is_blank_line:
            search_back_index -= 1

        pre_prev_token = actual_tokens[search_back_index]
        if pre_prev_token.is_end_token:
            end_token = cast(EndMarkdownToken, pre_prev_token)
            assert end_token.start_markdown_token, "End token must have a start token."
            pre_prev_token = end_token.start_markdown_token
            POGGER.debug(">>end_>using_start>>$", pre_prev_token)

        current_check = (
            # current_token.is_block and not current_token.is_link_reference_definition
            current_token.is_block
            or current_token.is_link_reference_definition
        )
        pre_prev_check = (
            pre_prev_token.is_block  # and not pre_prev_token.is_link_reference_definition
        )

        POGGER.debug(">>other--stack_count>>$", stack_count)
        POGGER.debug(
            ">>other--current_token>>$>>$",
            current_token,
            current_check,
        )
        POGGER.debug(
            ">>other--current_token-2>>$>>$",
            pre_prev_token,
            pre_prev_check,
        )
        return stack_count == 0 and current_check and pre_prev_check

    @staticmethod
    def __is_token_loose(
        actual_tokens: List[MarkdownToken], current_token_index: int, xx: bool = False
    ) -> bool:
        """
        Check to see if this token inspires looseness.
        """

        check_index = current_token_index - 1
        while actual_tokens[check_index].is_link_reference_definition:
            check_index -= 1
        token_to_check = actual_tokens[check_index]

        POGGER.debug("token_to_check-->$", token_to_check)
        is_loose = False
        if token_to_check.is_blank_line:
            POGGER.debug("before_blank-->$", actual_tokens[check_index - 1])
            if (
                actual_tokens[check_index - 1].is_new_list_item
                or actual_tokens[check_index - 1].is_list_start
            ):
                POGGER.debug("!!!Starting Blank!!!")
            elif actual_tokens[check_index - 1].is_block_quote_start and xx:
                POGGER.debug("!!!Starting BQ Blank!!!")
            else:
                POGGER.debug("!!!LOOSE!!!")
                is_loose = True
        return is_loose and TransformToGfmListLooseness.__is_really_loose(
            actual_tokens, check_index
        )

    @staticmethod
    def __is_really_loose(actual_tokens: List[MarkdownToken], check_index: int) -> bool:
        real_answer = None
        inner_containers = 0
        search_index = check_index - 1
        while real_answer is None and search_index >= 0:
            rt = actual_tokens[search_index]
            if rt.is_block_quote_end or rt.is_list_end:
                inner_containers += 1
            elif rt.is_list_start:
                if inner_containers:
                    inner_containers -= 1
                else:
                    real_answer = True
            elif rt.is_block_quote_start:
                if inner_containers:
                    inner_containers -= 1
                else:
                    real_answer = False
            search_index -= 1
        assert real_answer is not None, "must always have a real answer"
        return real_answer

    @staticmethod
    def __find_owning_list_start(
        actual_tokens: List[MarkdownToken], actual_token_index: int
    ) -> int:
        """
        Figure out what the list start for the current token is.
        """

        assert not actual_tokens[
            actual_token_index
        ].is_list_start, "Must not be a list start."
        current_index, stack_count = actual_token_index - 1, 0
        while True:
            assert current_index >= 0, "Must not go below 0."
            if actual_tokens[current_index].is_list_start:
                if stack_count == 0:
                    break
                stack_count -= 1
            elif actual_tokens[current_index].is_list_end:
                stack_count += 1
            current_index -= 1
        return current_index

    @staticmethod
    def reset_list_looseness(
        actual_tokens: List[MarkdownToken], actual_token_index: int
    ) -> bool:
        """
        Based on where we are within the actual tokens being emitted, figure
        out the correct list looseness to use.
        """

        POGGER.debug("!!!!!!!!!!!!!!!$", actual_token_index)
        search_index, stack_count, actual_tokens_size = (
            actual_token_index + 1,
            0,
            len(actual_tokens),
        )
        while search_index < actual_tokens_size:
            POGGER.debug(
                "!!$::$::$",
                stack_count,
                search_index,
                actual_tokens[search_index],
            )
            if actual_tokens[search_index].is_list_start:
                stack_count += 1
            elif actual_tokens[search_index].is_list_end:
                if not stack_count:
                    break
                stack_count -= 1
            search_index += 1
        POGGER.debug("!!!!!!!!!!!!!!!$-of-$", search_index, actual_tokens_size)
        # check to see where we are, then grab the matching start to find
        # the loose
        is_in_loose_list = search_index == actual_tokens_size
        if not is_in_loose_list:
            POGGER.debug(">>reset_list_looseness-token_list_start>>")
            new_index = TransformToGfmListLooseness.__find_owning_list_start(
                actual_tokens, search_index
            )
            POGGER.debug(">>reset_list_looseness>>$", new_index)
            list_token = cast(ListStartMarkdownToken, actual_tokens[new_index])
            is_in_loose_list = list_token.is_loose
        POGGER.debug("           is_in_loose_list=$", is_in_loose_list)
        return is_in_loose_list



"""
Module to encapuslate information about the current state of block quotes.
"""

from dataclasses import dataclass


@dataclass(frozen=True)
class BlockQuoteData:
    """
    Class to encapuslate information about the current state of block quotes.
    """

    current_count: int
    stack_count: int

    def __str__(self) -> str:
        return f"BlockQuoteData(current_count={self.current_count},stack_count={self.stack_count})"



"""
Module to provide for extending what is considered to be an autolink.
"""

from typing import Optional, Tuple, cast

from pymarkdown.container_blocks.parse_block_pass_properties import (
    ParseBlockPassProperties,
)
from pymarkdown.extension_manager.extension_impl import ExtensionDetails
from pymarkdown.extension_manager.extension_manager_constants import (
    ExtensionManagerConstants,
)
from pymarkdown.extension_manager.parser_extension import ParserExtension
from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.inline.inline_request import InlineRequest
from pymarkdown.inline.inline_response import InlineResponse
from pymarkdown.my_application_properties_facade import MyApplicationPropertiesFacade
from pymarkdown.tokens.email_autolink_markdown_token import EmailAutolinkMarkdownToken
from pymarkdown.tokens.text_markdown_token import TextMarkdownToken
from pymarkdown.tokens.uri_autolink_markdown_token import UriAutolinkMarkdownToken


class MarkdownExtendedAutolinksExtension(ParserExtension):
    """
    Extension to implement the extended autolinks extension.
    """

    __valid_autolink_domain_characters = (
        "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789_-"
    )
    __email_first_part = (
        "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789.-_+"
    )
    __email_domain_part = (
        "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789.-_"
    )
    __xmpp_resource_part = (
        "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789.@"
    )

    def get_identifier(self) -> str:
        """
        Get the identifier associated with this extension.
        """
        return "markdown-extended-autolinks"

    def get_details(self) -> ExtensionDetails:
        """
        Get the details for the extension.
        """
        return ExtensionDetails(
            extension_id=self.get_identifier(),
            extension_name="Markdown Extended Autolinks",
            extension_description="Allows extended parsing of Markdown Autolinks.",
            extension_enabled_by_default=False,
            extension_version="0.5.0",
            extension_interface_version=ExtensionManagerConstants.EXTENSION_INTERFACE_VERSION_BASIC,
            extension_url="https://pymarkdown.readthedocs.io/en/latest/extensions/extended-autolinks/",
            extension_configuration=None,
        )

    def apply_configuration(
        self, extension_specific_facade: MyApplicationPropertiesFacade
    ) -> None:
        """
        Apply any configuration required by the extension.
        """
        _ = extension_specific_facade  # pragma: no cover

    @staticmethod
    def __handle_www_autolink_prefix(
        inline_request: InlineRequest,
    ) -> Tuple[bool, bool, Optional[int]]:
        is_valid_prefix = not inline_request.next_index or inline_request.source_text[
            inline_request.next_index - 1
        ] in (" ", "(", "*", "_", "~")

        is_valid_start = False
        num_collected, new_index = ParserHelper.collect_while_character(
            inline_request.source_text, inline_request.next_index, "w"
        )
        if new_index is not None and new_index < len(inline_request.source_text):
            is_valid_start = (
                new_index is not None
                and inline_request.source_text[new_index] == "."
                and num_collected == 3
            )
            if is_valid_start:
                new_index += 1

        return is_valid_prefix, is_valid_start, new_index

    @staticmethod
    def __handle_www_autolink_domain(
        source_text: str, new_index: int
    ) -> Tuple[int, Optional[str]]:
        domain_parts = []
        keep_collecting = True
        while keep_collecting:
            index_after_collect, ex_text = ParserHelper.collect_while_one_of_characters(
                source_text,
                new_index,
                MarkdownExtendedAutolinksExtension.__valid_autolink_domain_characters,
            )
            if not ex_text:
                new_index -= 1
                break
            assert index_after_collect is not None
            domain_parts.append(ex_text)
            new_index = index_after_collect
            keep_collecting = (
                index_after_collect < len(source_text)
                and source_text[index_after_collect] == "."
            )
            if keep_collecting:
                new_index += 1
        if (
            len(domain_parts) >= 2
            and "_" not in domain_parts[-1]
            and "_" not in domain_parts[-2]
        ):
            return new_index, ".".join(domain_parts)
        return -1, None

    @staticmethod
    def __handle_www_autolink_valid_path(source_text: str, new_index: int) -> int:
        current_index = new_index
        if current_index < len(source_text) and source_text[current_index] == "/":
            current_index, _ = ParserHelper.collect_until_one_of_characters_verified(
                source_text, current_index + 1, " \t\n<"
            )

            # Strip any trailing characters.
            while source_text[current_index - 1] in "?!.,:*_~":
                current_index -= 1

            # Balance out any parentheses.
            curent_path = source_text[new_index:current_index]
            keep_going = True
            while keep_going and curent_path.endswith(")"):
                current_path_open_count = curent_path.count("(")
                current_path_close_count = curent_path.count(")")
                if current_path_close_count > current_path_open_count:
                    current_index -= 1
                    curent_path = source_text[new_index:current_index]
                else:
                    keep_going = False

            if curent_path.endswith(";"):
                amper_index = curent_path.rfind("&")
                if amper_index != -1:
                    reference_text = curent_path[amper_index + 1 : -1]
                    if reference_text.isalnum():
                        current_index -= len(reference_text) + 2
        return current_index

    @staticmethod
    def handle_www_autolink(
        parser_properties: ParseBlockPassProperties, inline_request: InlineRequest
    ) -> InlineResponse:
        """
        Take care of the "www." versions, that start with a 'w'
        """
        _ = parser_properties
        inline_response = InlineResponse()
        new_token, between_brackets = None, None

        (
            is_valid_prefix,
            is_valid_start,
            new_index,
        ) = MarkdownExtendedAutolinksExtension.__handle_www_autolink_prefix(
            inline_request
        )
        if is_valid_prefix and is_valid_start:
            assert new_index is not None
            (
                next_index,
                _,
            ) = MarkdownExtendedAutolinksExtension.__handle_www_autolink_domain(
                inline_request.source_text, new_index
            )
            if next_index != -1:
                next_index = (
                    MarkdownExtendedAutolinksExtension.__handle_www_autolink_valid_path(
                        inline_request.source_text, next_index
                    )
                )
                auto_link_text = inline_request.source_text[
                    inline_request.next_index : next_index
                ]
            if next_index != -1:
                assert inline_request.line_number is not None
                assert inline_request.column_number is not None
                assert inline_request.remaining_line is not None
                new_token = UriAutolinkMarkdownToken(
                    auto_link_text,
                    inline_request.line_number,
                    inline_request.column_number + len(inline_request.remaining_line),
                    add_http_prefix=True,
                )

        if new_token:
            (
                inline_response.new_string,
                inline_response.new_index,
                inline_response.new_tokens,
                between_brackets,
            ) = (
                "",
                next_index,
                [new_token],
                auto_link_text,
            )
        else:
            # Skip the current character
            inline_response.new_string, inline_response.new_index, between_brackets = (
                inline_request.source_text[inline_request.next_index : new_index],
                new_index,
                inline_request.source_text[inline_request.next_index : new_index],
            )

        (
            inline_response.delta_line_number,
            inline_response.delta_column_number,
        ) = ParserHelper.calculate_deltas(between_brackets)
        return inline_response

    @staticmethod
    def __handle_http_autolink_prefix(
        inline_request: InlineRequest,
    ) -> Tuple[bool, bool, int]:
        is_valid_prefix = not inline_request.next_index or inline_request.source_text[
            inline_request.next_index - 1
        ] in (" ", "(", "*", "_", "~")

        is_valid_start = False
        new_index = inline_request.next_index + 1
        first_pattern = "http://"
        second_pattern = "https://"
        rest_of_line = inline_request.source_text[inline_request.next_index :]
        if rest_of_line.startswith(first_pattern):
            is_valid_start = True
            new_index = inline_request.next_index + len(first_pattern)
        elif rest_of_line.startswith(second_pattern):
            is_valid_start = True
            new_index = inline_request.next_index + len(second_pattern)

        return is_valid_prefix, is_valid_start, new_index

    @staticmethod
    def handle_http_autolink(
        parser_properties: ParseBlockPassProperties, inline_request: InlineRequest
    ) -> InlineResponse:
        """
        Take care of the "http://" and "https://" versions, that start with an 'h'
        """
        _ = parser_properties
        inline_response = InlineResponse()
        new_token, between_brackets = None, None

        (
            is_valid_prefix,
            is_valid_start,
            new_index,
        ) = MarkdownExtendedAutolinksExtension.__handle_http_autolink_prefix(
            inline_request
        )
        if is_valid_prefix and is_valid_start:
            (
                next_index,
                _,
            ) = MarkdownExtendedAutolinksExtension.__handle_www_autolink_domain(
                inline_request.source_text, new_index
            )
            if next_index != -1:
                next_index = (
                    MarkdownExtendedAutolinksExtension.__handle_www_autolink_valid_path(
                        inline_request.source_text, next_index
                    )
                )
                autolink_text = inline_request.source_text[
                    inline_request.next_index : next_index
                ]
            if next_index != -1:
                assert inline_request.line_number is not None
                assert inline_request.column_number is not None
                assert inline_request.remaining_line is not None
                new_token = UriAutolinkMarkdownToken(
                    autolink_text,
                    inline_request.line_number,
                    inline_request.column_number + len(inline_request.remaining_line),
                    add_http_prefix=False,
                    add_angle_brackets=False,
                )

        if new_token:
            (
                inline_response.new_string,
                inline_response.new_index,
                inline_response.new_tokens,
                between_brackets,
            ) = (
                "",
                next_index,
                [new_token],
                autolink_text,
            )
        else:
            # Skip the current character
            inline_response.new_string, inline_response.new_index, between_brackets = (
                inline_request.source_text[inline_request.next_index : new_index],
                new_index,
                inline_request.source_text[inline_request.next_index : new_index],
            )

        (
            inline_response.delta_line_number,
            inline_response.delta_column_number,
        ) = ParserHelper.calculate_deltas(between_brackets)
        return inline_response

    @staticmethod
    def __handle_email_autolink_consume_inner(
        inline_request: InlineRequest,
        inline_response: InlineResponse,
        user_part_index: int,
        domain_part_index: int,
        user_id_part: str,
    ) -> Tuple[str, int]:
        collected_text = inline_request.source_text[user_part_index:domain_part_index]
        assert inline_request.current_string is not None
        assert inline_request.remaining_line is not None
        combined_text = inline_request.current_string + inline_request.remaining_line
        if not combined_text.endswith(user_id_part):
            left_index = len(inline_request.inline_blocks) - 1
            while (
                left_index >= 0
                and (
                    inline_request.inline_blocks[left_index].is_text
                    or inline_request.inline_blocks[left_index].is_special_text
                )
                and (len(combined_text) < len(user_id_part))
            ):
                text_token = cast(
                    TextMarkdownToken, inline_request.inline_blocks[left_index]
                )
                combined_text = text_token.token_text + combined_text
                left_index -= 1
            assert combined_text.endswith(user_id_part)
        inline_response.reduce_remaining_line_by = len(user_id_part)
        return collected_text, domain_part_index

    # pylint: disable=too-many-arguments
    @staticmethod
    def __handle_email_autolink_consume_token(
        inline_request: InlineRequest,
        inline_response: InlineResponse,
        user_part_index: int,
        domain_part_index: int,
        user_id_part: str,
        domain_part: str,
    ) -> Tuple[EmailAutolinkMarkdownToken, Optional[str], int]:
        assert user_part_index is not None
        assert domain_part_index is not None
        (
            collected_text,
            domain_part_index,
        ) = MarkdownExtendedAutolinksExtension.__handle_email_autolink_consume_inner(
            inline_request,
            inline_response,
            user_part_index,
            domain_part_index,
            user_id_part,
        )

        assert inline_request.line_number is not None
        assert inline_request.column_number is not None
        assert inline_request.remaining_line is not None
        if inline_request.remaining_line.endswith(user_id_part):
            new_column_number = inline_request.column_number - (
                len(user_id_part) - len(inline_request.remaining_line)
            )
            between_brackets = collected_text
        else:
            assert user_id_part.endswith(inline_request.remaining_line)
            fred = user_id_part[: -len(inline_request.remaining_line)]
            new_column_number = (
                inline_request.column_number
                + len(inline_request.remaining_line)
                - len(user_id_part)
            )
            between_brackets = inline_request.remaining_line + domain_part
            if fred[-1] != "_":
                between_brackets += "@"

        new_token = EmailAutolinkMarkdownToken(
            collected_text,
            inline_request.line_number,
            new_column_number,
            add_angle_brackets=False,
        )
        next_index = domain_part_index
        return new_token, between_brackets, next_index

    # pylint: enable=too-many-arguments

    @staticmethod
    def __handle_email_autolink_consume(
        inline_request: InlineRequest,
        inline_response: InlineResponse,
        at_index: int,
        between_brackets: Optional[str],
    ) -> Tuple[Optional[EmailAutolinkMarkdownToken], Optional[str], int]:
        _, user_part_index = (
            ParserHelper.collect_backwards_while_one_of_characters_verified(
                inline_request.source_text,
                at_index - 1,
                MarkdownExtendedAutolinksExtension.__email_first_part,
            )
        )
        user_id_part = inline_request.source_text[user_part_index:at_index]
        domain_part_index, domain_part = (
            ParserHelper.collect_while_one_of_characters_verified(
                inline_request.source_text,
                at_index + 1,
                MarkdownExtendedAutolinksExtension.__email_domain_part,
            )
        )
        new_token = None
        next_index = -1
        collected_text = inline_request.source_text[at_index:domain_part_index]
        if collected_text.endswith("."):
            collected_text = collected_text[:-1]
            domain_part_index -= 1
        if (
            ".." not in collected_text
            and collected_text.count(".") >= 1
            and not collected_text.endswith("-")
            and not collected_text.endswith("_")
        ):
            (
                new_token,
                between_brackets,
                next_index,
            ) = MarkdownExtendedAutolinksExtension.__handle_email_autolink_consume_token(
                inline_request,
                inline_response,
                user_part_index,
                domain_part_index,
                user_id_part,
                domain_part,
            )
        return new_token, between_brackets, next_index

    @staticmethod
    def __handle_mail_autolink_prefix(
        inline_request: InlineRequest,
    ) -> Tuple[bool, bool, int]:
        is_valid_prefix = not inline_request.next_index or inline_request.source_text[
            inline_request.next_index - 1
        ] in (" ", "(", "*", "_", "~")

        is_valid_start = False
        new_index = inline_request.next_index + 1
        first_pattern = "mailto:"
        second_pattern = "xmpp:"
        rest_of_line = inline_request.source_text[inline_request.next_index :]
        if rest_of_line.startswith(first_pattern):
            is_valid_start = True
            new_index = inline_request.next_index + len(first_pattern)
        elif rest_of_line.startswith(second_pattern):
            is_valid_start = True
            new_index = inline_request.next_index + len(second_pattern)

        return is_valid_prefix, is_valid_start, new_index

    @staticmethod
    def __handle_email_autolink_named_resource(
        inline_request: InlineRequest,
        is_valid_prefix: bool,
        is_valid_start: bool,
        domain_part_index: Optional[int],
    ) -> Optional[int]:
        if (
            is_valid_prefix
            and is_valid_start
            and inline_request.source_text[inline_request.next_index] == "x"
            and domain_part_index is not None
            and ParserHelper.is_character_at_index(
                inline_request.source_text, domain_part_index, "/"
            )
        ):
            after_ex_text_index, ex_text = ParserHelper.collect_while_one_of_characters(
                inline_request.source_text,
                domain_part_index + 1,
                MarkdownExtendedAutolinksExtension.__xmpp_resource_part,
            )
            if ex_text:
                domain_part_index = after_ex_text_index
        return domain_part_index

    @staticmethod
    def __handle_email_autolink_named_main(
        inline_request: InlineRequest,
    ) -> Tuple[bool, bool, int, int, Optional[int]]:
        next_index = -1
        collected_text = None
        domain_part_index: Optional[int] = None
        (
            is_valid_prefix,
            is_valid_start,
            new_index,
        ) = MarkdownExtendedAutolinksExtension.__handle_mail_autolink_prefix(
            inline_request
        )
        if is_valid_prefix and is_valid_start:
            user_part_index, _ = ParserHelper.collect_while_one_of_characters_verified(
                inline_request.source_text,
                new_index,
                MarkdownExtendedAutolinksExtension.__email_first_part,
            )
            is_valid_prefix = ParserHelper.is_character_at_index(
                inline_request.source_text, user_part_index, "@"
            )
        if is_valid_prefix and is_valid_start:
            user_part_index += 1
            domain_part_index, _ = (
                ParserHelper.collect_while_one_of_characters_verified(
                    inline_request.source_text,
                    user_part_index,
                    MarkdownExtendedAutolinksExtension.__email_domain_part,
                )
            )
            collected_text = inline_request.source_text[
                user_part_index:domain_part_index
            ]
            if collected_text.endswith("."):
                collected_text = collected_text[:-1]
                domain_part_index -= 1
            is_valid_prefix = (
                ".." not in collected_text
                and collected_text.count(".") >= 1
                and not collected_text.endswith("-")
                and not collected_text.endswith("_")
            )
        return (
            is_valid_prefix,
            is_valid_start,
            new_index,
            next_index,
            domain_part_index,
        )

    @staticmethod
    def __handle_email_autolink_named(
        parser_properties: ParseBlockPassProperties, inline_request: InlineRequest
    ) -> InlineResponse:
        """
        Take care of the "www." versions, that start with a 'w'
        """
        _ = parser_properties
        inline_response = InlineResponse()
        new_token, between_brackets = None, None
        domain_part_index: Optional[int] = None

        (
            is_valid_prefix,
            is_valid_start,
            new_index,
            next_index,
            domain_part_index,
        ) = MarkdownExtendedAutolinksExtension.__handle_email_autolink_named_main(
            inline_request
        )
        domain_part_index = (
            MarkdownExtendedAutolinksExtension.__handle_email_autolink_named_resource(
                inline_request, is_valid_prefix, is_valid_start, domain_part_index
            )
        )
        if is_valid_prefix and is_valid_start:
            assert domain_part_index is not None
            next_index = domain_part_index
            autolink_text = inline_request.source_text[
                inline_request.next_index : next_index
            ]
            assert inline_request.line_number is not None
            assert inline_request.column_number is not None
            new_token = UriAutolinkMarkdownToken(
                autolink_text,
                inline_request.line_number,
                inline_request.column_number,
                add_http_prefix=False,
                add_angle_brackets=False,
            )

        if new_token:
            (
                inline_response.new_string,
                inline_response.new_index,
                inline_response.new_tokens,
                between_brackets,
            ) = (
                "",
                next_index,
                [new_token],
                autolink_text,
            )
        else:
            # Skip the current character
            inline_response.new_string, inline_response.new_index, between_brackets = (
                inline_request.source_text[inline_request.next_index : new_index],
                new_index,
                inline_request.source_text[inline_request.next_index : new_index],
            )

        (
            inline_response.delta_line_number,
            inline_response.delta_column_number,
        ) = ParserHelper.calculate_deltas(between_brackets)
        return inline_response

    @staticmethod
    def handle_email_autolink(
        parser_properties: ParseBlockPassProperties, inline_request: InlineRequest
    ) -> InlineResponse:
        """
        Handle a normal email autolink
        """
        ff = inline_request.source_text[inline_request.next_index]
        if ff != "@":
            return MarkdownExtendedAutolinksExtension.__handle_email_autolink_named(
                parser_properties, inline_request
            )
        _ = parser_properties
        inline_response = InlineResponse()
        new_token = None
        between_brackets: Optional[str] = None
        new_index = inline_request.next_index + 1
        next_index = inline_request.next_index

        at_index = inline_request.next_index
        if 0 < at_index < len(inline_request.source_text):
            (
                new_token,
                between_brackets,
                next_index,
            ) = MarkdownExtendedAutolinksExtension.__handle_email_autolink_consume(
                inline_request, inline_response, at_index, between_brackets
            )

        if new_token:
            (
                inline_response.new_string,
                inline_response.new_index,
                inline_response.new_tokens,
            ) = (
                "",
                next_index,
                [new_token],
            )
        else:
            # Skip the current character
            inline_response.new_string, inline_response.new_index, between_brackets = (
                inline_request.source_text[inline_request.next_index : new_index],
                new_index,
                inline_request.source_text[inline_request.next_index : new_index],
            )

        assert between_brackets is not None
        (
            inline_response.delta_line_number,
            inline_response.delta_column_number,
        ) = ParserHelper.calculate_deltas(between_brackets)
        return inline_response



"""
Module to implement a plugin that looks for inconsistencies in the
style used for Unordered List elements.
"""

from typing import Dict, List, cast

from pymarkdown.plugin_manager.plugin_details import (
    PluginDetailsV2,
    PluginDetailsV3,
    QueryConfigItem,
)
from pymarkdown.plugin_manager.plugin_scan_context import PluginScanContext
from pymarkdown.plugin_manager.rule_plugin import RulePlugin
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.tokens.unordered_list_start_markdown_token import (
    UnorderedListStartMarkdownToken,
)


class RuleMd004(RulePlugin):
    """
    Class to implement a plugin that looks for inconsistencies in the
    style used for Unordered List elements.
    """

    __consistent_style = "consistent"
    __asterisk_style = "asterisk"
    __plus_style = "plus"
    __dash_style = "dash"
    __sublist_style = "sublist"

    __valid_styles = [
        __consistent_style,
        __asterisk_style,
        __plus_style,
        __dash_style,
        __sublist_style,
    ]

    def __init__(self) -> None:
        super().__init__()
        self.__style_type = ""
        self.__actual_style_type: Dict[int, str] = {}
        self.__current_list_level = 0

    def get_details(self) -> PluginDetailsV2:
        """
        Get the details for the plugin.
        """
        return PluginDetailsV3(
            # bullet, ul
            plugin_name="ul-style",
            plugin_id="MD004",
            plugin_enabled_by_default=True,
            plugin_description="Inconsistent Unordered List Start style",
            plugin_version="0.6.0",
            plugin_url="https://pymarkdown.readthedocs.io/en/latest/plugins/rule_md004.md",
            plugin_configuration="style",
            plugin_supports_fix=True,
        )

    @classmethod
    def __validate_configuration_style(cls, found_value: str) -> None:
        if found_value not in RuleMd004.__valid_styles:
            raise ValueError(f"Allowable values: {RuleMd004.__valid_styles}")

    def initialize_from_config(self) -> None:
        """
        Event to allow the plugin to load configuration information.
        """
        self.__style_type = self.plugin_configuration.get_string_property_with_default(
            "style",
            RuleMd004.__consistent_style,
            valid_value_fn=self.__validate_configuration_style,
        )

    def query_config(self) -> List[QueryConfigItem]:
        """
        Query to find out the configuration that the rule is using.
        """
        return [QueryConfigItem("style", self.__style_type)]

    def starting_new_file(self) -> None:
        """
        Event that the a new file to be scanned is starting.
        """
        self.__actual_style_type = {}
        self.__current_list_level = 0
        if self.__style_type not in (
            RuleMd004.__consistent_style,
            RuleMd004.__sublist_style,
        ):
            self.__actual_style_type[0] = self.__style_type

    @classmethod
    def __get_sequence_type(cls, token: UnorderedListStartMarkdownToken) -> str:
        if token.list_start_sequence == "*":
            return RuleMd004.__asterisk_style
        if token.list_start_sequence == "+":
            return RuleMd004.__plus_style
        assert token.list_start_sequence == "-"
        return RuleMd004.__dash_style

    def __next_token_triggered(
        self, context: PluginScanContext, token: MarkdownToken, this_start_style: str
    ) -> None:
        if context.in_fix_mode:
            if (
                self.__actual_style_type[self.__current_list_level]
                == RuleMd004.__plus_style
            ):
                new_start_sequence = "+"
            elif (
                self.__actual_style_type[self.__current_list_level]
                == RuleMd004.__dash_style
            ):
                new_start_sequence = "-"
            else:
                new_start_sequence = "*"
            self.register_fix_token_request(
                context, token, "next_token", "list_start_sequence", new_start_sequence
            )
        else:
            extra_data = (
                f"Expected: {self.__actual_style_type[self.__current_list_level]}; "
                + f"Actual: {this_start_style}"
            )
            self.report_next_token_error(context, token, extra_data)

    def next_token(self, context: PluginScanContext, token: MarkdownToken) -> None:
        """
        Event that a new token is being processed.
        """
        if token.is_unordered_list_start:
            list_token = cast(UnorderedListStartMarkdownToken, token)
            if self.__current_list_level not in self.__actual_style_type:
                if self.__style_type in (RuleMd004.__sublist_style,) or (
                    self.__style_type in (RuleMd004.__consistent_style)
                    and not self.__actual_style_type
                ):
                    self.__actual_style_type[self.__current_list_level] = (
                        self.__get_sequence_type(list_token)
                    )
                else:
                    self.__actual_style_type[self.__current_list_level] = (
                        self.__actual_style_type[0]
                    )

            this_start_style = self.__get_sequence_type(list_token)
            if self.__actual_style_type[self.__current_list_level] != this_start_style:
                self.__next_token_triggered(context, token, this_start_style)
            self.__current_list_level += 1
        elif token.is_unordered_list_end:
            self.__current_list_level -= 1



"""
Module to implement a sample plugin that just reports that it has been called.
"""

from pymarkdown.plugin_manager.plugin_details import PluginDetails
from pymarkdown.plugin_manager.plugin_scan_context import PluginScanContext
from pymarkdown.plugin_manager.rule_plugin import RulePlugin
from pymarkdown.tokens.markdown_token import MarkdownToken


class PluginTestException(Exception):
    """
    Custom exception for the PluginOne class.
    """

    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(self.message)


class PluginOne(RulePlugin):
    """
    Class to implement a sample plugin that just reports that it has been called.
    """

    __valid_values = [0, 1, 2]

    def __init__(self) -> None:
        super().__init__()
        self.test_value: int = -1
        self.other_test_value: int = -1

    def get_details(self) -> PluginDetails:
        """
        Get the details for the plugin.
        """
        return PluginDetails(
            plugin_name="debug-only",
            plugin_id="MD999",
            plugin_enabled_by_default=False,
            plugin_description="Debug plugin",
            plugin_version="0.0.0",
            plugin_interface_version=1,
        )

    @classmethod
    def __validate_configuration_other_test_value(cls, found_value: int) -> None:
        if found_value not in PluginOne.__valid_values:
            raise ValueError(f"Allowable values: {found_value}")

    def initialize_from_config(self) -> None:
        """
        Event to allow the plugin to load configuration.
        """
        print(f"{self.get_details().plugin_id}>>init_from_config")
        self.test_value = self.plugin_configuration.get_integer_property_with_default(
            "test_value", 1
        )
        print(f"{self.get_details().plugin_id}>>test_value>>{self.test_value}")
        self.other_test_value = (
            self.plugin_configuration.get_integer_property_with_default(
                "other_test_value",
                1,
                valid_value_fn=self.__validate_configuration_other_test_value,
            )
        )
        print(
            f"{self.get_details().plugin_id}>>other_test_value>>{self.other_test_value}"
        )
        if self.test_value == 10:
            raise PluginTestException("because")

    def starting_new_file(self) -> None:
        """
        Event that the a new file to be scanned is starting.
        """
        print(f"{self.get_details().plugin_id}>>starting_new_file>>")

    def next_line(self, context: PluginScanContext, line: str) -> None:
        """
        Event that a new line is being processed.
        """
        _ = context
        print(f"{self.get_details().plugin_id}>>next_line:{line}")

    def next_token(self, context: PluginScanContext, token: MarkdownToken) -> None:
        """
        Event that a new token is being processed.
        """
        _ = context
        print(f"{self.get_details().plugin_id}>>token:{token}")
        if self.test_value == 20:
            raise PluginTestException("because")

    def completed_file(self, context: PluginScanContext) -> None:
        """
        Event that the file being currently scanned is now completed.
        """
        _ = context
        print(f"{self.get_details().plugin_id}>>completed_file")



"""
Module to provide processing for indented leaf blocks.
"""

import logging
from typing import List, Optional, Tuple, cast

from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.general.parser_state import ParserState
from pymarkdown.general.position_marker import PositionMarker
from pymarkdown.general.tab_helper import TabHelper
from pymarkdown.tokens.block_quote_markdown_token import BlockQuoteMarkdownToken
from pymarkdown.tokens.indented_code_block_markdown_token import (
    IndentedCodeBlockMarkdownToken,
)
from pymarkdown.tokens.list_start_markdown_token import ListStartMarkdownToken
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.tokens.stack_token import IndentedCodeBlockStackToken, ListStackToken
from pymarkdown.tokens.text_markdown_token import TextMarkdownToken

POGGER = ParserLogger(logging.getLogger(__name__))


# pylint: disable=too-few-public-methods


class IndentedLeafBlockProcessor:
    """
    Class to provide processing for indented leaf blocks.
    """

    # pylint: disable=too-many-arguments, too-many-locals
    @staticmethod
    def parse_indented_code_block(
        parser_state: ParserState,
        position_marker: PositionMarker,
        extracted_whitespace: str,
        removed_chars_at_start: int,
        last_block_quote_index: int,
        original_line: str,
    ) -> List[MarkdownToken]:
        """
        Handle the parsing of an indented code block
        """

        new_tokens: List[MarkdownToken] = []
        if (
            TabHelper.is_length_greater_than_or_equal_to(
                extracted_whitespace, 4, start_index=removed_chars_at_start
            )
            and not parser_state.token_stack[-1].is_paragraph
        ):
            # POGGER.debug(
            #     "parse_indented_code_block>>start",
            # )
            # POGGER.debug("removed_chars_at_start>:$:<", removed_chars_at_start)
            # POGGER.debug("extracted_whitespace>:$:<", extracted_whitespace)
            # POGGER.debug(
            #     "position_marker.text_to_parse>:$:<", position_marker.text_to_parse
            # )
            # POGGER.debug(
            #     "position_marker.text_to_parse[index_number=$]>:$:<",
            #     position_marker.index_number,
            #     position_marker.text_to_parse[position_marker.index_number :],
            # )
            # POGGER.debug("original_line>:$:<", original_line)

            indented_text = position_marker.text_to_parse[
                position_marker.index_number :
            ]

            last_block_stack_index = parser_state.find_last_block_quote_on_stack()
            is_in_block_quote = last_block_stack_index > 0
            POGGER.debug("is_in_block_quote>:$:<", is_in_block_quote)

            last_list_stack_index = parser_state.find_last_list_block_on_stack()
            is_in_list = last_list_stack_index > 0
            POGGER.debug("is_in_list>:$:<", is_in_list)

            tabified_extracted_space: Optional[str] = None
            xx_extracted_space, xx_left_over, adjust_block_quote_indent = (
                None,
                None,
                False,
            )
            if ParserHelper.tab_character in original_line:
                (
                    tabified_extracted_space,
                    xx_extracted_space,
                    xx_left_over,
                    adjust_block_quote_indent,
                ) = IndentedLeafBlockProcessor.__parse_indented_code_block_with_tab(
                    parser_state,
                    position_marker,
                    is_in_list,
                    is_in_block_quote,
                    original_line,
                    last_block_stack_index,
                    last_list_stack_index,
                )

            IndentedLeafBlockProcessor.__process_indented_code_block(
                parser_state,
                position_marker,
                new_tokens,
                tabified_extracted_space,
                original_line,
                xx_extracted_space,
                xx_left_over,
                adjust_block_quote_indent,
                last_block_quote_index,
                extracted_whitespace,
                indented_text,
            )
        else:
            POGGER.debug(
                "parse_indented_code_block>>not eligible",
            )
        return new_tokens

    # pylint: enable=too-many-arguments, too-many-locals

    @staticmethod
    def __parse_indented_code_block_with_tab_list_list(
        parser_state: ParserState, original_line: str, last_list_index: int
    ) -> Tuple[ListStartMarkdownToken, str, int, bool, int]:
        last_list_token = cast(
            ListStartMarkdownToken,
            parser_state.token_stack[last_list_index].matching_markdown_token,
        )
        last_list_lead_spaces = last_list_token.leading_spaces
        is_list_start_line = False
        list_tabbed_adjust = -1
        if last_list_lead_spaces is not None:
            POGGER.debug("last_list_lead_spaces>:$:<", last_list_lead_spaces)
            POGGER.debug("original_line>:$:<", original_line)
            lead_space_last_line_index = last_list_lead_spaces.rfind("\n")
            if lead_space_last_line_index != -1:
                last_list_lead_spaces = last_list_lead_spaces[
                    lead_space_last_line_index + 1 :
                ]
                POGGER.debug("last_block_quote_lead_spaces>:$:<", last_list_lead_spaces)
        else:
            is_list_start_line = True
            list_tabbed_adjust = last_list_token.tabbed_adjust
            last_list_lead_spaces = (
                last_list_token.list_start_content + last_list_token.list_start_sequence
                if last_list_token.is_ordered_list_start
                else last_list_token.list_start_sequence
            )
            xx_delta = last_list_token.indent_level - len(last_list_lead_spaces)
            last_list_lead_spaces += " " * xx_delta

        lead_space_len = len(last_list_lead_spaces)
        POGGER.debug("lead_space_len>:$:<", lead_space_len)
        return (
            last_list_token,
            last_list_lead_spaces,
            lead_space_len,
            is_list_start_line,
            list_tabbed_adjust,
        )

    # pylint: disable=too-many-arguments
    @staticmethod
    def __parse_indented_code_block_with_tab_list_calc(
        detabbed_ex_part: str,
        is_list_start_line: bool,
        ex_part: str,
        list_tabbed_adjust: int,
        fex_space: str,
        indent_used: int,
        search_index: int,
    ) -> Tuple[str, str]:
        xx_left_over = ""
        delta = len(detabbed_ex_part) - 4
        if is_list_start_line:
            xx_left_over = " " * delta
            if delta != 0:
                xx_left_over = ParserHelper.create_replacement_markers(
                    ParserHelper.replace_noop_character, xx_left_over
                )
            if (
                ex_part
                and ex_part[0] == "\t"
                and list_tabbed_adjust >= 0
                and delta != 0
            ):
                ex_part = ex_part[1:]
        elif ex_part[-1] == "\t" and delta != 0:
            xx_left_over = " " * delta
            xx_left_over = ParserHelper.create_replacement_markers(
                ParserHelper.replace_noop_character, xx_left_over
            )
        xx_left_over += fex_space[indent_used + search_index :]
        return xx_left_over, ex_part

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-locals
    @staticmethod
    def __parse_indented_code_block_with_tab_list(
        parser_state: ParserState,
        position_marker: PositionMarker,
        original_line: str,
        last_list_index: int,
    ) -> Tuple[Optional[str], str, str, bool]:
        (
            last_list_token,
            last_list_lead_spaces,
            lead_space_len,
            is_list_start_line,
            list_tabbed_adjust,
        ) = IndentedLeafBlockProcessor.__parse_indented_code_block_with_tab_list_list(
            parser_state, original_line, last_list_index
        )

        _, ex_space = ParserHelper.extract_spaces_verified(
            position_marker.text_to_parse, 0
        )
        # POGGER.debug("after_space_index>:$:<", after_space_index)
        # POGGER.debug("ex_space>:$:<", ex_space)
        assert (
            len(ex_space) >= 4
        ), "An indented code block must be indented by at least 4 spaces."
        was_indented = not parser_state.token_stack[-2].is_document
        fex_space, fex_space_index, split_tab = TabHelper.find_tabified_string(
            original_line,
            position_marker.text_to_parse,
            use_proper_traverse=False,
            reconstruct_prefix=last_list_lead_spaces,
            was_indented=was_indented,
        )
        assert original_line.endswith(
            fex_space
        ), "Original line must end with the tabified string."

        indent_used = 0
        extra_index = 0
        if split_tab:
            extra_index = lead_space_len
            search_index, _, _ = TabHelper.search_for_tabbed_prefix2(
                fex_space, fex_space_index, 0, 0, lead_space_len
            )
            indent_used = search_index - 1

        (
            search_index,
            ex_part,
            detabbed_ex_part,
        ) = TabHelper.search_for_tabbed_prefix2(
            fex_space, fex_space_index, extra_index, indent_used, 4
        )

        (
            xx_left_over,
            ex_part,
        ) = IndentedLeafBlockProcessor.__parse_indented_code_block_with_tab_list_calc(
            detabbed_ex_part,
            is_list_start_line,
            ex_part,
            list_tabbed_adjust,
            fex_space,
            indent_used,
            search_index,
        )

        if not is_list_start_line:
            xx_dd = (
                original_line[:indent_used]
                if split_tab
                else original_line[: -(len(fex_space))]
            )
            POGGER.debug(
                "__parse_indented_code_block_with_tab_list>>last_list_token>>$",
                last_list_token,
            )
            last_list_token.remove_last_leading_space()
            POGGER.debug(
                "__parse_indented_code_block_with_tab_list>>last_list_token>>$",
                last_list_token,
            )
            POGGER.debug(
                "__parse_indented_code_block_with_tab_list>>list_token>>$",
                last_list_token,
            )
            last_list_token.add_leading_spaces(xx_dd)
            POGGER.debug(
                "__parse_indented_code_block_with_tab_list>>list_token>>$",
                last_list_token,
            )
        return None, ex_part, xx_left_over, False

    # pylint: enable=too-many-locals

    # pylint: disable=too-many-arguments, too-many-locals
    @staticmethod
    def __parse_indented_code_block_with_tab(
        parser_state: ParserState,
        position_marker: PositionMarker,
        is_in_list: bool,
        is_in_block_quote: bool,
        original_line: str,
        last_block_quote_index: int,
        last_list_index: int,
    ) -> Tuple[Optional[str], Optional[str], Optional[str], bool]:
        if not is_in_list and not is_in_block_quote:
            next_character = position_marker.text_to_parse[position_marker.index_number]
            next_character_index = original_line.find(next_character)
            assert (
                next_character_index != -1
            ), "Next character must be found in the original line."
            return (
                original_line[:next_character_index],
                None,
                None,
                False,
            )

        # Note not handling lists yet
        # assert is_in_block_quote

        if last_list_index > last_block_quote_index:
            return IndentedLeafBlockProcessor.__parse_indented_code_block_with_tab_list(
                parser_state, position_marker, original_line, last_list_index
            )
        (
            last_block_quote_lead_spaces,
            adj_lead_space_len,
            adjust_block_quote_indent,
            other_adjustment,
        ) = IndentedLeafBlockProcessor.__get_indented_block_with_tab_quote_properties(
            parser_state, last_block_quote_index, original_line
        )

        lead_space_len = len(last_block_quote_lead_spaces)
        POGGER.debug("lead_space_len>:$:<", lead_space_len)
        after_space_index, ex_space = ParserHelper.extract_spaces_verified(
            original_line, lead_space_len + adj_lead_space_len
        )
        return IndentedLeafBlockProcessor.__parse_indented_code_block_with_tab_complete(
            after_space_index,
            ex_space,
            lead_space_len + other_adjustment,
            original_line,
            adjust_block_quote_indent,
        )

    # pylint: enable=too-many-arguments, too-many-locals

    @staticmethod
    def __parse_indented_code_block_with_tab_complete(
        after_space_index: int,
        ex_space: str,
        lead_space_len: int,
        original_line: str,
        adjust_block_quote_indent: bool,
    ) -> Tuple[Optional[str], str, str, bool]:
        POGGER.debug("after_space_index>:$:<", after_space_index)
        POGGER.debug("ex_space>:$:<", ex_space)
        detabified_ex_space = TabHelper.detabify_string(
            ex_space, additional_start_delta=lead_space_len
        )
        POGGER.debug("detabified_ex_space>:$:<", detabified_ex_space)

        (last_good_space_index, space_prefix, detabified_ex_space) = (
            TabHelper.search_for_tabbed_prefix(ex_space, lead_space_len, 4)
        )

        if len(detabified_ex_space) == 4:
            xx_extracted_space = space_prefix
            xx_left_over = ex_space[last_good_space_index:]
        else:
            xx_extracted_space = space_prefix[:-1]
            xx_left_over = ParserHelper.create_replacement_markers(
                space_prefix[-1], detabified_ex_space[4:]
            )

        POGGER.debug("xx_extracted_space>:$:<", xx_extracted_space)
        POGGER.debug("xx_left_over>:$:<", xx_left_over)
        xx_left_over += original_line[after_space_index:]
        POGGER.debug("xx_left_over>:$:<", xx_left_over)
        return None, xx_extracted_space, xx_left_over, adjust_block_quote_indent

    # pylint: disable=too-many-arguments
    @staticmethod
    def __process_indented_code_block(
        parser_state: ParserState,
        position_marker: PositionMarker,
        new_tokens: List[MarkdownToken],
        tabified_extracted_space: Optional[str],
        original_line: str,
        xx_extracted_space: Optional[str],
        xx_left_over: Optional[str],
        adjust_block_quote_indent: bool,
        last_block_quote_index: int,
        extracted_whitespace: str,
        indented_text: str,
    ) -> None:
        # adj_ws: Optional[str] = ""
        # POGGER.debug("tabified_extracted_space>:$:<", tabified_extracted_space)
        if not parser_state.token_stack[-1].is_indented_code_block:
            # POGGER.debug("xx_extracted_space>:$:<", xx_extracted_space)
            # POGGER.debug("xx_left_over>:$:<", xx_left_over)
            (
                last_block_quote_index,
                extracted_whitespace,
                indented_text,
            ) = IndentedLeafBlockProcessor.__create_indented_block(
                parser_state,
                last_block_quote_index,
                position_marker,
                extracted_whitespace,
                new_tokens,
                tabified_extracted_space,
                original_line,
                indented_text,
                xx_extracted_space,
                xx_left_over,
            )
        elif tabified_extracted_space:
            (_, adj_ws, _) = IndentedLeafBlockProcessor.__recalculate_whitespace(
                extracted_whitespace, 0, tabified_extracted_space
            )
            extracted_whitespace = adj_ws

            adj_ws_length = len(adj_ws)
            indented_text = original_line[adj_ws_length:]
        elif xx_extracted_space is not None:
            extracted_whitespace = xx_extracted_space
            assert (
                xx_left_over is not None
            ), "If we extracted spaces, there must be some left over."
            indented_text = xx_left_over

        # POGGER.debug("extracted_whitespace>:$:<", extracted_whitespace)
        # POGGER.debug("indented_text>:$:<", indented_text)

        assert not adjust_block_quote_indent
        # if adjust_block_quote_indent:
        #     TabHelper.adjust_block_quote_indent_for_tab(parser_state)

        new_tokens.append(
            TextMarkdownToken(
                indented_text,
                extracted_whitespace,
                position_marker=position_marker,
            )
        )

    # pylint: enable=too-many-arguments

    @staticmethod
    def __get_indented_block_with_tab_quote_properties(
        parser_state: ParserState, last_block_quote_index: int, original_line: str
    ) -> Tuple[str, int, bool, int]:
        adjust_block_quote_indent = False

        POGGER.debug(
            "token_stack[last_block_quote_index]>:$:<",
            parser_state.token_stack[last_block_quote_index],
        )
        last_block_quote_token = cast(
            BlockQuoteMarkdownToken,
            parser_state.token_stack[last_block_quote_index].matching_markdown_token,
        )
        last_block_quote_lead_spaces = last_block_quote_token.bleading_spaces
        assert (
            last_block_quote_lead_spaces is not None
        ), "Bleading spaces must be defined by now."
        POGGER.debug("last_block_quote_lead_spaces>:$:<", last_block_quote_lead_spaces)
        lead_space_last_line_index = last_block_quote_lead_spaces.rfind("\n")
        POGGER.debug("original_line>:$:<", original_line)
        adj_lead_space_len = 0
        if lead_space_last_line_index != -1:
            last_block_quote_lead_spaces = last_block_quote_lead_spaces[
                lead_space_last_line_index + 1 :
            ]
            POGGER.debug(
                "last_block_quote_lead_spaces>:$:<", last_block_quote_lead_spaces
            )
        if last_block_quote_lead_spaces[-1] == " ":
            last_part_minus_tailing_space = last_block_quote_lead_spaces[:-1]
        else:
            last_part_minus_tailing_space = last_block_quote_lead_spaces
            return (
                last_block_quote_lead_spaces,
                adj_lead_space_len,
                adjust_block_quote_indent,
                1,
            )
        POGGER.debug(
            "last_part_minus_tailing_space>:$:<", last_part_minus_tailing_space
        )
        assert original_line.startswith(
            last_part_minus_tailing_space
        ), "Original line must start with the last split part, minus any tailing space."
        trailing_char_in_original = original_line[len(last_part_minus_tailing_space)]
        POGGER.debug("trailing_char_in_original>:$:<", trailing_char_in_original)
        assert trailing_char_in_original != ParserHelper.tab_character
        # if trailing_char_in_original == ParserHelper.tab_character:
        #     adjust_block_quote_indent = True
        #     adj_lead_space_len = -1
        POGGER.debug("last_block_quote_lead_spaces>:$:<", last_block_quote_lead_spaces)
        # if not adjust_block_quote_indent:
        #     assert original_line.startswith(last_block_quote_lead_spaces)
        return (
            last_block_quote_lead_spaces,
            adj_lead_space_len,
            adjust_block_quote_indent,
            0,
        )

    # pylint: disable=too-many-arguments
    @staticmethod
    def __create_indented_block(
        parser_state: ParserState,
        last_block_quote_index: int,
        position_marker: PositionMarker,
        extracted_whitespace: str,
        new_tokens: List[MarkdownToken],
        tabified_extracted_space: Optional[str],
        original_line: str,
        indented_text: str,
        xx_extracted_space: Optional[str],
        xx_left_over: Optional[str],
    ) -> Tuple[int, str, str]:
        column_number = (
            position_marker.index_number
            + position_marker.index_indent
            - len(extracted_whitespace)
            + 1
        )
        POGGER.debug("last_block_quote_index>:$:<", last_block_quote_index)
        POGGER.debug("extracted_whitespace>:$:<", extracted_whitespace)
        POGGER.debug("tabified_extracted_space>:$:<", tabified_extracted_space)
        POGGER.debug("xx_extracted_space>:$:<", xx_extracted_space)
        (
            last_block_quote_index,
            actual_whitespace_index,
            adj_ws,
            extracted_whitespace,
        ) = IndentedLeafBlockProcessor.__prepare_for_indented_block(
            parser_state,
            last_block_quote_index,
            extracted_whitespace,
            tabified_extracted_space,
        )
        POGGER.debug("last_block_quote_index>:$:<", last_block_quote_index)
        POGGER.debug("actual_whitespace_index>:$:<", actual_whitespace_index)
        POGGER.debug("adj_ws>:$:<", adj_ws)
        POGGER.debug("extracted_whitespace>:$:<", extracted_whitespace)
        if xx_extracted_space is not None:
            adj_ws = xx_extracted_space
        POGGER.debug("adj_ws>:$:<", adj_ws)

        column_number += actual_whitespace_index
        POGGER.debug("column_number>>$", column_number)

        new_token = IndentedCodeBlockMarkdownToken(
            adj_ws, position_marker.line_number, column_number
        )
        parser_state.token_stack.append(IndentedCodeBlockStackToken(new_token))
        new_tokens.append(new_token)
        POGGER.debug("left_ws>>$<<", extracted_whitespace)

        POGGER.debug("xx_left_over>:$:<", xx_left_over)
        POGGER.debug("tabified_extracted_space>:$:<", tabified_extracted_space)
        if xx_left_over:
            extracted_whitespace = ""
            indented_text = xx_left_over
        if tabified_extracted_space:
            extracted_whitespace = ""
            adj_ws_length = len(adj_ws)
            indented_text = original_line[adj_ws_length:]

        POGGER.debug("extracted_whitespace>:$:<", extracted_whitespace)
        POGGER.debug("indented_text>:$:<", indented_text)
        return last_block_quote_index, extracted_whitespace, indented_text

    # pylint: enable=too-many-arguments

    @staticmethod
    def __prepare_for_indented_block(
        parser_state: ParserState,
        last_block_quote_index: int,
        extracted_whitespace: str,
        tabified_extracted_space: Optional[str],
    ) -> Tuple[int, int, str, str]:
        POGGER.debug(">>>>$", parser_state.token_stack[-1])
        if parser_state.token_stack[-1].is_list:
            list_token = cast(ListStackToken, parser_state.token_stack[-1])
            POGGER.debug(
                ">>indent>>$",
                list_token.indent_level,
            )
            last_block_quote_index = 0

        POGGER.debug(
            "__recalculate_whitespace>>$>>$",
            extracted_whitespace,
            0,
        )
        (
            actual_whitespace_index,
            adj_ws,
            left_ws,
        ) = IndentedLeafBlockProcessor.__recalculate_whitespace(
            extracted_whitespace, 0, tabified_extracted_space
        )
        return (
            last_block_quote_index,
            actual_whitespace_index,
            adj_ws,
            left_ws,
        )

    @staticmethod
    def __recalculate_whitespace(
        whitespace_to_parse: str,
        offset_index: int,
        tabified_extracted_space: Optional[str],
    ) -> Tuple[int, str, str]:
        """
        Recalculate the whitespace characteristics.
        """
        POGGER.debug("whitespace_to_parse>>$>>", whitespace_to_parse)
        POGGER.debug("offset_index>>$>>", offset_index)
        POGGER.debug("tabified_extracted_space>>$>>", tabified_extracted_space)

        actual_whitespace_index = 4 + offset_index
        adj_ws = whitespace_to_parse[:actual_whitespace_index]
        left_ws = whitespace_to_parse[actual_whitespace_index:]
        if tabified_extracted_space:
            additional_start_delta = 0
            length_so_far = 0
            last_index = 0
            next_character_index = 0
            while (
                next_character_index < len(tabified_extracted_space)
                and length_so_far < 4
            ):
                next_character = tabified_extracted_space[next_character_index]
                POGGER.debug("next_character>:$:<", next_character)
                length_so_far = (
                    (1 + (length_so_far // 4)) * 4
                    if next_character == ParserHelper.tab_character
                    else length_so_far + 1
                )
                last_index += 1
                POGGER.debug("length_so_far>:$:<", length_so_far)
                next_character_index += 1
            POGGER.debug("length_so_far>:$:<", length_so_far)
            assert (
                length_so_far == 4
            ), "length_so_far must equate to a prefix of exactly 4 characters."
            POGGER.debug("last_index>:$:<", last_index)
            tabbed_prefix = tabified_extracted_space[:last_index]
            POGGER.debug("tabbed_prefix>:$:<", tabbed_prefix)
            err = TabHelper.detabify_string(
                tabbed_prefix,
                additional_start_delta=additional_start_delta,
            )
            POGGER.debug("tabbed_prefix>:$:<", tabbed_prefix)
            POGGER.debug("err>:$:<", err)
            assert (
                len(err) == 4
            ), "Tabbed prefix must equate to a prefix of exactly 4 characters."
            adj_ws = tabbed_prefix
            left_ws = tabified_extracted_space[last_index:]

        POGGER.debug("actual_whitespace_index>>$", actual_whitespace_index)
        POGGER.debug("adj_ws>>$<<", adj_ws)
        POGGER.debug("left_ws>>$<<", left_ws)
        return actual_whitespace_index, adj_ws, left_ws


# pylint: enable=too-few-public-methods



"""
Module to provide for an encapsulation of the text element.
"""

import logging
from typing import List, Optional, Tuple, Union, cast

from typing_extensions import override

from pymarkdown.general.constants import Constants
from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.general.position_marker import PositionMarker
from pymarkdown.tokens.indented_code_block_markdown_token import (
    IndentedCodeBlockMarkdownToken,
)
from pymarkdown.tokens.inline_markdown_token import InlineMarkdownToken
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.tokens.paragraph_markdown_token import ParagraphMarkdownToken
from pymarkdown.transform_gfm.transform_state import TransformState
from pymarkdown.transform_markdown.markdown_transform_context import (
    MarkdownTransformContext,
    RegisterHtmlTransformHandlersProtocol,
    RegisterMarkdownTransformHandlersProtocol,
)

POGGER = ParserLogger(logging.getLogger(__name__))


class TextMarkdownToken(InlineMarkdownToken):
    """
    Class to provide for an encapsulation of the text element.
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        token_text: str,
        extracted_whitespace: str,
        end_whitespace: Optional[str] = None,
        position_marker: Optional[PositionMarker] = None,
        line_number: int = 0,
        column_number: int = 0,
        is_special: bool = False,
        tabified_text: Optional[str] = None,
    ):
        (
            self.__token_text,
            self.__extracted_whitespace,
            self.__end_whitespace,
            self.__tabified_text,
        ) = (token_text, extracted_whitespace, end_whitespace, tabified_text)
        InlineMarkdownToken.__init__(
            self,
            MarkdownToken._token_text,
            "",
            position_marker=position_marker,
            line_number=line_number,
            column_number=column_number,
            is_special=is_special,
        )
        self.__compose_extra_data_field()

    # pylint: enable=too-many-arguments
    # pylint: disable=protected-access
    @staticmethod
    def get_markdown_token_type() -> str:
        """
        Get the type of markdown token for rehydration purposes.
        """
        return MarkdownToken._token_text

    # pylint: enable=protected-access

    def _set_token_text(self, new_text: str) -> None:
        self.__token_text = new_text
        self.__compose_extra_data_field()

    @property
    def token_text(self) -> str:
        """
        Returns the text associated with the token.
        """
        return self.__token_text

    @property
    def extracted_whitespace(self) -> str:
        """
        Returns any whitespace that was extracted before the processing of this element occurred.
        """
        return self.__extracted_whitespace

    @property
    def end_whitespace(self) -> Optional[str]:
        """
        Returns any whitespace that was extracted after the processing of this element occurred.
        """
        return self.__end_whitespace

    @property
    def tabified_text(self) -> Optional[str]:
        """
        Returns any text that had a tab character in it.
        """
        return self.__tabified_text

    def create_copy(self) -> "TextMarkdownToken":
        """
        Create a copy of this token.
        """
        return TextMarkdownToken(
            self.__token_text,
            self.__extracted_whitespace,
            self.__end_whitespace,
            line_number=self.line_number,
            column_number=self.column_number,
        )

    @override
    def _modify_token(self, field_name: str, field_value: Union[str, int]) -> bool:
        if field_name == "token_text" and isinstance(field_value, str):
            self.__token_text = field_value
            self.__compose_extra_data_field()

            return True
        if field_name == "end_whitespace" and isinstance(field_value, str):
            self.__end_whitespace = field_value
            self.__compose_extra_data_field()

            return True
        if field_name == "extracted_whitespace" and isinstance(field_value, str):
            self.__extracted_whitespace = field_value
            self.__compose_extra_data_field()
            return True
        return False

    def __compose_extra_data_field(self) -> None:
        """
        Compose the object's self.extra_data field from the local object's variables.
        """

        data_field_parts = [self.__token_text, self.__extracted_whitespace]
        if self.__end_whitespace:
            data_field_parts.append(self.__end_whitespace)
        elif self.__tabified_text:
            data_field_parts.extend(("", self.__tabified_text))
        self._set_extra_data(MarkdownToken.extra_data_separator.join(data_field_parts))

    def remove_final_whitespace(self) -> str:
        """
        Remove any final whitespace.  Used by paragraph blocks so that they do not
        end with a hard break.
        """

        removed_whitespace = ""
        # POGGER.debug("self.__tabified_text=:$:", self.__tabified_text)
        # POGGER.debug("self.__token_text=:$:", self.__token_text)
        (
            collected_whitespace_length,
            first_non_whitespace_index,
        ) = ParserHelper.collect_backwards_while_one_of_characters_verified(
            self.__token_text, -1, Constants.ascii_whitespace
        )
        # POGGER.debug("collected_whitespace_length=:$:", collected_whitespace_length)
        # POGGER.debug("first_non_whitespace_index=:$:", first_non_whitespace_index)

        if collected_whitespace_length:
            removed_whitespace = self.__token_text[
                first_non_whitespace_index : first_non_whitespace_index
                + collected_whitespace_length
            ]
            self.__token_text = self.__token_text[:first_non_whitespace_index]
            if self.__tabified_text:
                (
                    collected_whitespace_length,
                    first_non_whitespace_index,
                ) = ParserHelper.collect_backwards_while_one_of_characters_verified(
                    self.__tabified_text, -1, Constants.ascii_whitespace
                )
                # POGGER.debug("collected_whitespace_length=:$:", collected_whitespace_length)
                # POGGER.debug("first_non_whitespace_index=:$:", first_non_whitespace_index)
                removed_whitespace = self.__tabified_text[
                    first_non_whitespace_index : first_non_whitespace_index
                    + collected_whitespace_length
                ]
                self.__tabified_text = self.__tabified_text[:first_non_whitespace_index]
        self.__compose_extra_data_field()
        return removed_whitespace

    def combine(
        self,
        other_text_token: MarkdownToken,
        remove_leading_spaces: int,
        only_change_text_blocks: bool,
    ) -> str:
        """
        Combine the two text tokens together with a line feed between.
        If remove_leading_spaces > 0, then that many leading spaces will be
        removed from each line, if present.
        If remove_leading_spaces == -1, then.
        If remove_leading_spaces == 0, then.
        """

        # if self.line_number == other_text_token.line_number:
        if only_change_text_blocks:

            return self.__combine_only_text_blocks(other_text_token)
        if other_text_token.is_blank_line:
            text_to_combine = ""
            tabified_text_to_combine: Optional[str] = ""
            assert (
                other_text_token.extra_data is not None
            ), "If the other token is a blank line, the extra_data field must be set."
            (
                whitespace_present,
                blank_line_sequence,
            ) = (
                other_text_token.extra_data,
                ParserHelper.replace_noop_character,
            )
        else:
            assert other_text_token.is_text, "if not a blank line, must be a text token"
            text_other_token = cast(TextMarkdownToken, other_text_token)

            text_to_combine = text_other_token.token_text
            tabified_text_to_combine = text_other_token.tabified_text
            (
                whitespace_present,
                blank_line_sequence,
            ) = (
                text_other_token.extracted_whitespace,
                "",
            )

        removed_whitespace, prefix_whitespace = self.__combine_handle_whitespace(
            remove_leading_spaces, whitespace_present
        )

        if self.__tabified_text or tabified_text_to_combine:
            other_token_text = tabified_text_to_combine or text_to_combine

            this_token_text = self.__tabified_text or self.__token_text
            self.__tabified_text = (
                f"{this_token_text}{ParserHelper.newline_character}{blank_line_sequence}"
                + f"{prefix_whitespace}{other_token_text}"
            )
        self.__token_text = (
            f"{self.__token_text}{ParserHelper.newline_character}{blank_line_sequence}"
            + f"{prefix_whitespace}{text_to_combine}"
        )
        self.__compose_extra_data_field()
        return removed_whitespace

    def __combine_only_text_blocks(self, other_text_token: MarkdownToken) -> str:
        the_other_text_token = cast(TextMarkdownToken, other_text_token)
        assert the_other_text_token is not None
        self.__token_text = self.token_text + the_other_text_token.token_text
        self.__extracted_whitespace = (
            self.extracted_whitespace + the_other_text_token.extracted_whitespace
        )
        if the_other_text_token.end_whitespace is not None:
            self.__end_whitespace = (
                (self.end_whitespace + the_other_text_token.end_whitespace)
                if self.end_whitespace is not None
                else the_other_text_token.end_whitespace
            )
        self.__compose_extra_data_field()
        return ""

    def __combine_handle_whitespace(
        self, remove_leading_spaces: int, whitespace_present: str
    ) -> Tuple[str, str]:
        prefix_whitespace = ""
        removed_whitespace = ""

        if not remove_leading_spaces:
            prefix_whitespace = whitespace_present
        elif remove_leading_spaces == -1:
            prefix_whitespace = ""
            self.__extracted_whitespace = (
                f"{self.__extracted_whitespace}"
                + f"{ParserHelper.newline_character}{whitespace_present}"
            )
        else:
            whitespace_present_size = len(whitespace_present)
            POGGER.debug(
                "whitespace_present>>$>>$<<",
                whitespace_present_size,
                whitespace_present,
            )
            POGGER.debug("remove_leading_spaces>>$<<", remove_leading_spaces)
            removed_whitespace = (
                whitespace_present
                if whitespace_present_size < remove_leading_spaces
                else whitespace_present[:remove_leading_spaces]
            )
            prefix_whitespace = (
                ""
                if whitespace_present_size < remove_leading_spaces
                else whitespace_present[remove_leading_spaces:]
            )
        return removed_whitespace, prefix_whitespace

    def register_for_markdown_transform(
        self, registration_function: RegisterMarkdownTransformHandlersProtocol
    ) -> None:
        """
        Register any rehydration handlers for leaf markdown tokens.
        """
        registration_function(
            TextMarkdownToken,
            TextMarkdownToken.__rehydrate_text,
            None,
        )

    @staticmethod
    def __rehydrate_text(
        context: MarkdownTransformContext,
        current_token: MarkdownToken,
        previous_token: Optional[MarkdownToken],
    ) -> str:
        """
        Rehydrate the text from the token.
        """
        _ = previous_token

        if (
            context.block_stack[-1].is_inline_link
            or context.block_stack[-1].is_inline_image
        ):
            return ""

        prefix_text = ""
        current_text_token = cast(TextMarkdownToken, current_token)
        POGGER.debug(
            f">>rehydrate_text>>:{ParserHelper.make_value_visible(current_text_token.token_text)}:<<"
        )
        # main_text = ParserHelper.resolve_noops_from_text(current_text_token.token_text)
        main_text = ParserHelper.remove_all_from_text(
            current_text_token.token_text, include_noops=True
        )

        POGGER.debug(f"<<rehydrate_text>>{ParserHelper.make_value_visible(main_text)}")

        POGGER.debug(
            f">>leading_whitespace>>:{ParserHelper.make_value_visible(current_text_token.extracted_whitespace)}:<<"
        )
        leading_whitespace = ParserHelper.remove_all_from_text(
            current_text_token.extracted_whitespace
        )
        POGGER.debug(
            f"<<leading_whitespace>>:{ParserHelper.make_value_visible(leading_whitespace)}:<<"
        )

        extra_line = ""
        # assert context.block_stack
        if context.block_stack[-1].is_indented_code_block:
            code_block_token = cast(
                IndentedCodeBlockMarkdownToken, context.block_stack[-1]
            )
            (
                main_text,
                prefix_text,
                leading_whitespace,
            ) = TextMarkdownToken.__reconstitute_indented_text(
                main_text,
                code_block_token.extracted_whitespace,
                code_block_token.indented_whitespace,
                leading_whitespace,
            )
        elif context.block_stack[-1].is_html_block:
            extra_line = ParserHelper.newline_character
        elif context.block_stack[-1].is_paragraph:
            main_text = TextMarkdownToken.__reconstitute_paragraph_text(
                context, main_text, current_text_token
            )
        elif context.block_stack[-1].is_setext_heading:
            main_text = TextMarkdownToken.__reconstitute_setext_text(
                main_text, current_text_token
            )

        POGGER.debug(
            f"<<prefix_text>>{ParserHelper.make_value_visible(prefix_text)}"
            + f"<<leading_whitespace>>{ParserHelper.make_value_visible(leading_whitespace)}"
            + f"<<main_text>>{ParserHelper.make_value_visible(main_text)}<<"
        )
        return "".join([prefix_text, leading_whitespace, main_text, extra_line])

    @staticmethod
    def __reconstitute_paragraph_text(
        context: MarkdownTransformContext,
        main_text: str,
        current_token: "TextMarkdownToken",
    ) -> str:
        """
        For a paragraph block, figure out the text that got us here.
        """
        if ParserHelper.newline_character in main_text:
            owner_paragraph_token = cast(
                ParagraphMarkdownToken, context.block_stack[-1]
            )
            (
                main_text,
                owner_paragraph_token.rehydrate_index,
            ) = ParserHelper.recombine_string_with_whitespace(
                main_text,
                owner_paragraph_token.extracted_whitespace,
                owner_paragraph_token.rehydrate_index,
            )
            assert (
                current_token.end_whitespace
            ), "if there is a newline, there must be end_whitespace."
            main_text, _ = ParserHelper.recombine_string_with_whitespace(
                main_text,
                current_token.end_whitespace,
                start_text_index=0,
                post_increment_index=True,
                add_whitespace_after=True,
            )
        return main_text

    @staticmethod
    def __reconstitute_indented_text(
        main_text: str,
        prefix_text: str,
        indented_whitespace: str,
        leading_whitespace: str,
    ) -> Tuple[str, str, str]:
        """
        For an indented code block, figure out the text that got us here.
        """
        recombined_text, _ = ParserHelper.recombine_string_with_whitespace(
            main_text,
            f"{prefix_text}{leading_whitespace}{indented_whitespace}",
            start_text_index=0,
            post_increment_index=True,
        )
        return f"{recombined_text}{ParserHelper.newline_character}", "", ""

    @staticmethod
    def __reconstitute_setext_text_item(
        text_part_index: int,
        text_part_value: str,
        rejoined_token_text: List[str],
        split_parent_whitespace_text: List[str],
    ) -> None:
        ws_prefix_text = ""
        ws_suffix_text = ""
        if split_parent_whitespace_text[text_part_index]:
            split_setext_text = split_parent_whitespace_text[text_part_index].split(
                ParserHelper.whitespace_split_character
            )
            split_setext_text_size = len(split_setext_text)
            if split_setext_text_size == 1:
                assert text_part_index == 0, "This must match with the line below."
                ws_suffix_text = split_setext_text[0]
                # if text_part_index == 0:
                #     ws_suffix_text = split_setext_text[0]
                # else:
                #     ws_prefix_text = split_setext_text[0]
            else:
                assert split_setext_text_size == 2, "if not 1 part, must be 2 parts."
                ws_prefix_text = split_setext_text[0]
                ws_suffix_text = split_setext_text[1]

        rejoined_token_text.append(
            "".join([ws_prefix_text, text_part_value, ws_suffix_text])
        )

    @staticmethod
    def __reconstitute_setext_text(
        main_text: str, current_token: "TextMarkdownToken"
    ) -> str:
        """
        For a setext heading block, figure out the text that got us here.

        Because of the unique formatting of the setext data, the recombine_string_with_whitespace
        function cannot be used for this.
        """

        if ParserHelper.newline_character in main_text:
            split_token_text = main_text.split(ParserHelper.newline_character)
            assert (
                current_token.end_whitespace is not None
            ), "if there is a newline, there must be end_whitespace."
            split_parent_whitespace_text = current_token.end_whitespace.split(
                ParserHelper.newline_character
            )

            rejoined_token_text: List[str] = []
            for text_part_index, text_part_value in enumerate(split_token_text):
                TextMarkdownToken.__reconstitute_setext_text_item(
                    text_part_index,
                    text_part_value,
                    rejoined_token_text,
                    split_parent_whitespace_text,
                )

            main_text = ParserHelper.newline_character.join(rejoined_token_text)
        else:
            POGGER.debug(f"main_text>>{ParserHelper.make_value_visible(main_text)}")
            POGGER.debug(
                f"current_token>>{ParserHelper.make_value_visible(current_token)}"
            )
            if current_token.end_whitespace and current_token.end_whitespace.endswith(
                ParserHelper.whitespace_split_character
            ):
                main_text = f"{current_token.end_whitespace[:-1]}{main_text}"
        return main_text

    @staticmethod
    def register_for_html_transform(
        register_handlers: RegisterHtmlTransformHandlersProtocol,
    ) -> None:
        """
        Register any functions required to generate HTML from the tokens.
        """
        register_handlers(
            TextMarkdownToken, TextMarkdownToken.__handle_text_token, None
        )

    @staticmethod
    def __handle_text_token(
        output_html: str,
        next_token: MarkdownToken,
        transform_state: TransformState,
    ) -> str:
        """
        Handle the text token.
        """
        text_token = cast(TextMarkdownToken, next_token)
        adjusted_text_token = ParserHelper.resolve_all_from_text(text_token.token_text)

        token_parts: List[str] = []
        if transform_state.is_in_code_block:
            POGGER.debug(
                "text_token.extracted_whitespace>:$:<", text_token.extracted_whitespace
            )
            leading_space = ParserHelper.resolve_all_from_text(
                text_token.extracted_whitespace
            )

            POGGER.debug("leading_space>:$:<", leading_space)
            POGGER.debug("adjusted_text_token>:$:<", adjusted_text_token)
            token_parts.extend(
                [
                    leading_space,
                    adjusted_text_token,
                ]
            )
        elif transform_state.is_in_html_block:
            POGGER.debug(
                "text_token.extracted_whitespace>:$:<", text_token.extracted_whitespace
            )
            leading_space = ParserHelper.resolve_all_from_text(
                text_token.extracted_whitespace
            )

            POGGER.debug("leading_space>:$:<", leading_space)
            POGGER.debug("adjusted_text_token>:$:<", adjusted_text_token)
            POGGER.debug("newline_character>:$:<", ParserHelper.newline_character)
            token_parts.extend(
                [
                    leading_space,
                    adjusted_text_token,
                    ParserHelper.newline_character,
                ]
            )
        elif transform_state.is_in_setext_block:
            token_parts.append(adjusted_text_token)
        else:
            TextMarkdownToken.__handle_text_token_normal(
                token_parts, text_token, adjusted_text_token
            )

        token_parts.insert(0, output_html)
        return "".join(token_parts)

    @staticmethod
    def __handle_text_token_normal(
        token_parts: List[str],
        text_token: "TextMarkdownToken",
        adjusted_text_token: str,
    ) -> None:
        POGGER.debug("adjusted_text_token>:$:<", adjusted_text_token)
        POGGER.debug("text_token.end_whitespace>:$:<", text_token.end_whitespace)
        if text_token.end_whitespace is not None:
            newlines_in_adjusted = ParserHelper.count_newlines_in_text(
                adjusted_text_token
            )

            resolved_whitespace = ParserHelper.resolve_all_from_text(
                text_token.end_whitespace
            )
            POGGER.debug("resolved_whitespace>:$:<", resolved_whitespace)
            newlines_in_whitespace = ParserHelper.count_newlines_in_text(
                resolved_whitespace
            )
            arrays_to_combine: List[List[str]] = []
            if newlines_in_adjusted == newlines_in_whitespace:
                arrays_to_combine.append(
                    adjusted_text_token.split(ParserHelper.newline_character)
                )
            else:
                TextMarkdownToken.__handle_text_token_normal_enhanced(
                    arrays_to_combine, text_token
                )

            arrays_to_combine.append(
                resolved_whitespace.split(ParserHelper.newline_character)
            )
            assert len(arrays_to_combine[0]) == len(
                arrays_to_combine[1]
            ), "Items must have the same length."
            POGGER.debug("arrays_to_combine>:$:<", arrays_to_combine)
            final_parts: List[str] = []
            for loop_index in range(len(arrays_to_combine[0]) * 2):
                loop_index_mod = loop_index % 2
                loop_index_div = loop_index // 2
                if loop_index_mod == 0 and final_parts:
                    final_parts.append("\n")
                POGGER.debug(
                    "$-->final_parts>:$:<...$,$",
                    loop_index,
                    final_parts,
                    loop_index_mod,
                    loop_index_div,
                )
                final_parts.append(arrays_to_combine[loop_index_mod][loop_index_div])
                POGGER.debug("final_parts>:$:<", final_parts)
            adjusted_text_token = "".join(final_parts)
        token_parts.append(adjusted_text_token)

    @staticmethod
    def __handle_text_token_normal_enhanced(
        arrays_to_combine: List[List[str]], text_token: "TextMarkdownToken"
    ) -> None:
        """
        This should only happen in the case where there are replacement characters that
        include replacing a character sequence with a newline.  Specifically target that.
        """

        assert (
            "\a" in text_token.token_text
        ), "Token_text must have at least one replace character to be here."
        replace_character_count = ParserHelper.count_characters_in_text(
            text_token.token_text, "\a"
        )
        assert (
            replace_character_count % 3 == 0
        ), "Replace characters always occur in threes."

        start_index = 0
        current_line = ""
        processed_lines: List[str] = []
        next_index, found_prefix = (
            ParserHelper.collect_until_one_of_characters_verified(
                text_token.token_text, start_index, "\a\n"
            )
        )
        # POGGER.debug("next_index>:$:<", next_index)
        # POGGER.debug("found_prefix>:$:<", found_prefix)
        while next_index < len(text_token.token_text):
            current_line += found_prefix
            if text_token.token_text[next_index] == "\a":
                # POGGER.debug("before>:$:<", text_token.token_text[start_index:next_index])
                middle_index = text_token.token_text.find("\a", next_index + 1)
                # POGGER.debug("replace>:$:<", text_token.token_text[next_index+1:middle_index])
                start_index = text_token.token_text.find("\a", middle_index + 1)
                # POGGER.debug("with>:$:<", text_token.token_text[middle_index+1:start_index])
                current_line += text_token.token_text[middle_index + 1 : start_index]
            else:
                processed_lines.append(current_line)
                current_line = ""
                start_index = next_index
            next_index, found_prefix = (
                ParserHelper.collect_until_one_of_characters_verified(
                    text_token.token_text, start_index + 1, "\a\n"
                )
            )
            # POGGER.debug("next_index>:$:<", next_index)
            # POGGER.debug("found_prefix>:$:<", found_prefix)
        current_line += found_prefix

        # POGGER.debug("currernt_line>:$:<", current_line)
        assert current_line, "current_line must contain some text."
        processed_lines.append(current_line)

        arrays_to_combine.append(processed_lines)
